###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module provides functions to extract distinct parts of the fleur xml files
for easy versioning and reuse
"""
from __future__ import annotations

from masci_tools.io.parsers.fleur_schema import schema_dict_version_dispatch
from masci_tools.io.common_functions import AtomSiteProperties
from masci_tools.util.typing import XMLLike
from masci_tools.io.parsers import fleur_schema
from masci_tools.io.fleur_xml import FleurXMLContext

from lxml import etree
import warnings
import numpy as np
from logging import Logger
from typing import Any

from .xpathbuilder import FilterType


def get_fleur_modes(xmltree: XMLLike,
                    schema_dict: fleur_schema.InputSchemaDict | fleur_schema.OutputSchemaDict,
                    logger: Logger | None = None) -> dict[str, Any]:
    """
    Determine the calculation modes of fleur for the given xml file. Calculation modes
    are things that change the produced files or output in the out.xml files

    :param xmltree: etree representing the fleur xml file
    :param schema_dict: schema dictionary corresponding to the file version
                        of the xmltree
    :param logger: logger object for logging warnings, errors

    :returns: dictionary with all the extracted calculation modes

    The following modes are inspected:

        - `jspin`: How many spins are considered in the calculation
        - `noco`: Is the calculation non-collinear?
        - `soc`: Is spin-orbit coupling included?
        - `relax`: Is the calculation a structure relaxation?
        - `gw`: Special mode for GW/Spex calculations
        - `force_theorem`: Is a Force theorem calculation performed?
        - `film`: Is the structure a film system
        - `ldau`: Is LDA+U included?
        - `dos`: Is it a density of states calculation?
        - `band`: Is it a bandstructure calculation?
        - `bz_integration`: How is the integration over the Brillouin-Zone performed?

    """

    fleur_modes = {}
    with FleurXMLContext(xmltree, schema_dict, logger=logger) as root:
        fleur_modes['jspin'] = root.attribute('jspins')
        fleur_modes['noco'] = root.attribute('l_noco', default=False)
        fleur_modes['soc'] = root.attribute('l_soc', default=False)
        fleur_modes['relax'] = root.attribute('l_f', default=False)
        fleur_modes['gw'] = False
        if root.tag_exists('expertModes'):
            gw = root.attribute('gw', optional=True)
            if gw is None and schema_dict.inp_version >= (0, 34):
                gw = root.attribute('spex', default=0)
            fleur_modes['gw'] = gw != 0

        if schema_dict.inp_version > (0, 27):
            fleur_modes['force_theorem'] = root.tag_exists('forceTheorem')
        else:
            fleur_modes['force_theorem'] = False

        fleur_modes['cf_coeff'] = False
        if schema_dict.inp_version >= (0, 33) and root.tag_exists('cFCoeffs'):
            cf_coeff = any(root.attribute('potential', tag_name='cFCoeffs', list_return=True, optional=True))
            cf_coeff = cf_coeff or any(
                root.attribute('chargeDensity', tag_name='cFCoeffs', list_return=True, optional=True))
            fleur_modes['cf_coeff'] = cf_coeff

        fleur_modes['plot'] = False
        if root.tag_exists('plotting'):
            plot = root.attribute('iplot', default=False)
            if schema_dict.inp_version >= (0, 29):
                plot = isinstance(plot, int) and plot != 0
            fleur_modes['plot'] = plot

        fleur_modes['film'] = root.tag_exists('filmPos')
        fleur_modes['ldau'] = root.tag_exists('ldaU', contains='species')
        fleur_modes['dos'] = root.attribute('dos')
        fleur_modes['band'] = root.attribute('band')
        fleur_modes['bz_integration'] = root.attribute('mode', tag_name='bzIntegration')

        greensf = False
        if schema_dict.inp_version >= (0, 32):
            #We make the assumption that the existence of a greensfCalculation
            #tag implies the existence of a greens function calculation
            greensf = root.tag_exists('greensfCalculation', contains='species')
            if schema_dict.inp_version >= (0, 35):
                greensf = greensf or root.tag_exists('torqueCalculation', contains='species')
            else:
                greensf = greensf or root.tag_exists('torgueCalculation', contains='species')
        fleur_modes['greensf'] = greensf

        ldahia = False
        if schema_dict.inp_version >= (0, 32):
            ldahia = root.tag_exists('ldaHIA', contains='species')
        fleur_modes['ldahia'] = ldahia

    return fleur_modes


@schema_dict_version_dispatch(output_schema=False)
def get_nkpts(xmltree: XMLLike,
              schema_dict: fleur_schema.InputSchemaDict | fleur_schema.OutputSchemaDict,
              logger: Logger | None = None) -> int:
    """
    Get the number of kpoints that will be used in the calculation specified in the given
    fleur XMl file.

    .. warning::
        For file versions before Max5 only kPointList or kPointCount tags will work. However,
        for kPointCount there is no real guarantee that for every occasion it will correspond
        to the number of kpoints. So a warning is written out

    :param xmltree: etree representing the fleur xml file
    :param schema_dict: schema dictionary corresponding to the file version
                        of the xmltree
    :param logger: logger object for logging warnings, errors

    :returns: int with the number of kpoints
    """
    with FleurXMLContext(xmltree, schema_dict, logger=logger) as root:
        #Get the name of the current selected kPointSet
        list_name = root.attribute('listName')
        all_names = set(
            root.attribute('name', tag_name='kPointList', contains='kPointLists', list_return=True, optional=True))
        if list_name not in all_names:
            raise ValueError(f'Selected Kpoint list with the name: {list_name} does not exist\n'
                             f'Available list names: {all_names}')

        with root.find('kPointList', contains='kPointLists', filters={'kPointList': {'name': list_name}}) as kpoints:
            nkpts = kpoints.attribute('count')

    if not isinstance(nkpts, int):
        raise ValueError('Failed to evaluate nkpts')

    return nkpts


@get_nkpts.register(max_version='0.31')
def get_nkpts_max4(xmltree: XMLLike,
                   schema_dict: fleur_schema.InputSchemaDict | fleur_schema.OutputSchemaDict,
                   logger: Logger | None = None) -> int:
    """
    Get the number of kpoints that will be used in the calculation specified in the given
    fleur XMl file. Version specific for Max4 versions or older

    .. warning::
        For file versions before Max5 only kPointList or kPointCount tags will work. However,
        for kPointCount there is no real guarantee that for every occasion it will correspond
        to the number of kpoints. So a warning is written out

    :param xmltree: etree representing the fleur xml file
    :param schema_dict: schema dictionary corresponding to the file version
                        of the xmltree
    :param logger: logger object for logging warnings, errors

    :returns: int with the number of kpoints
    """
    modes = get_fleur_modes(xmltree, schema_dict, logger=logger)
    with FleurXMLContext(xmltree, schema_dict, logger=logger) as root:

        nkpts = None
        if modes['band'] or modes['gw']:
            filters = {'altKPointSet': {'purpose': 'bands' if modes['band'] else 'gw'}}
            if root.tag_exists('altKPointSet', filters=filters):
                with root.find('altKPointSet', filters=filters) as kpoints:
                    nkpts = kpoints.attribute('count', tag_name='kPointList', optional=True)
                    if nkpts is None:
                        nkpts = kpoints.attribute('count', tag_name='kPointCount', optional=True)
                        if nkpts is not None:
                            warnings.warn('kPointCount is not guaranteed to result in the given number of kpoints')
                    if nkpts is None:
                        raise ValueError('No kPointList or kPointCount found')

        output_schema = getattr(schema_dict, 'out_version', None) is not None
        if nkpts is None and output_schema:
            nkpts = root.attribute('count', tag_name='kPointList', contains='numericalParameters')
        elif nkpts is None and not output_schema:
            nkpts = root.attribute('count',
                                   tag_name='kPointList',
                                   not_contains=['altKPointSet', 'numericalParameters'],
                                   optional=True)
            if nkpts is None:
                nkpts = root.attribute('count', tag_name='kPointCount', not_contains='altKPointSet', optional=True)
                if nkpts is not None:
                    warnings.warn('kPointCount is not guaranteed to result in the given number of kpoints')
            if nkpts is None:
                raise ValueError('No kPointList or kPointCount found')

    if not isinstance(nkpts, int):
        raise ValueError('Failed to evaluate nkpts')

    return nkpts


def get_cell(xmltree: XMLLike,
             schema_dict: fleur_schema.InputSchemaDict | fleur_schema.OutputSchemaDict,
             logger: Logger | None = None,
             convert_to_angstroem: bool = True) -> tuple[np.ndarray, tuple[bool, bool, bool]]:
    """
    Get the Bravais matrix from the given fleur xml file. In addition a list
    determining in, which directions there are periodic boundary conditions
    in the system.

    .. warning::
        Only the explicit definition of the Bravais matrix is supported.
        Old inputs containing the `latnam` definitions are not supported

    :param xmltree: etree representing the fleur xml file
    :param schema_dict: schema dictionary corresponding to the file version
                        of the xmltree
    :param logger: logger object for logging warnings, errors
    :param convert_to_angstroem: bool if True the bravais matrix is converted to angstroem

    :returns: numpy array of the bravais matrix and list of boolean values for
              periodic boundary conditions
    """
    from masci_tools.util.constants import BOHR_A

    NO_CELL_ERROR = 'Could not extract Bravais matrix out of inp.xml. Is the ' \
                    'Bravais matrix explicitly given? i.e Latnam definition ' \
                    'not supported.'

    cell: np.ndarray | None = None
    with FleurXMLContext(xmltree, schema_dict, logger=logger) as root:

        lattice_tag: etree._Element | None = None
        if root.tag_exists('bulkLattice'):
            lattice_tag = root.simple_xpath('bulkLattice')  #type:ignore[assignment]
            pbc = (True, True, True)
        elif root.tag_exists('filmLattice'):
            lattice_tag = root.simple_xpath('filmLattice')  #type:ignore[assignment]
            pbc = (True, True, False)

        if lattice_tag is None:
            raise ValueError(NO_CELL_ERROR)

        with root.nested(lattice_tag) as lattice:
            scale = lattice.attribute('scale', not_contains={'/a', 'c/'})

            bravais_tag: etree._Element
            if lattice.tag_exists('bravaisMatrix'):
                bravais_tag = lattice.simple_xpath('bravaisMatrix')  #type:ignore[assignment]
            elif not all(pbc) and schema_dict.inp_version >= (0, 35):
                bravais_tag = lattice.simple_xpath('bravaisMatrixFilm')  #type:ignore[assignment]
            else:
                raise ValueError(NO_CELL_ERROR)

            film_matrix = bravais_tag.tag == 'bravaisMatrixFilm'
            with lattice.nested(bravais_tag) as bravais:

                row1 = bravais.text('row-1', optional=True)
                row2 = bravais.text('row-2', optional=True)
                if film_matrix:
                    dtilda = lattice.attribute('dtilda')
                    row1 += [0.0]
                    row2 += [0.0]
                    row3 = [0.0, 0.0, dtilda]
                else:
                    row3 = bravais.text('row-3', optional=True)

            if all(x is not None and x != [] for x in [row1, row2, row3]):
                cell = np.array([row1, row2, row3]) * scale
                if convert_to_angstroem and cell is not None:
                    cell *= BOHR_A

    if cell is None:
        raise ValueError(NO_CELL_ERROR)

    return cell, pbc


def _get_species_info(xmltree: XMLLike,
                      schema_dict: fleur_schema.InputSchemaDict | fleur_schema.OutputSchemaDict,
                      logger: Logger | None = None) -> dict[str, dict[str, str]]:
    """
    Gets the species identifiers and information.
    Used to keep species information consistent between
    :py:func:`get_parameter_data` and :py:func:`get_structure_data`

    :param xmltree: etree representing the fleur xml file
    :param schema_dict: schema dictionary corresponding to the file version
                        of the xmltree
    :param logger: logger object for logging warnings, errors

    :returns: Tuple of dicts, containing the normalized species ids
              and the elements for each species
    """
    import re

    with FleurXMLContext(xmltree, schema_dict, logger=logger) as root:

        names = root.attribute('name', contains='species', list_return=True)
        elements = root.attribute('element', contains='species', list_return=True)

    if len(names) != len(elements):
        raise ValueError(
            f'Failed to read in species names and elements. Got {len(names)} names and {len(elements)} elements')

    species_info: dict[str, dict[str, str]] = {}
    for name, element in zip(names, elements):
        #Check if the species name has a numerical id at the end (separated by - or .)
        #And add all of them first
        species_info[name] = {}
        species_info[name]['element'] = element
        species_info[name]['normed_name'] = name
        match = re.fullmatch(r'(.+[\-\.])([1-9]+)', name)
        if match:
            species_info[name]['id'] = match.group(2)

    for name, info in species_info.items():
        if 'id' not in info:
            element = info['element']
            #Find the smallest id which is free
            used_ids = {
                int(val['id']) for name, val in species_info.items() if 'id' in val and val['element'] == element
            }
            possible_ids = range(1, max(used_ids, default=0) + 2)
            info['id'] = str(min(set(possible_ids) - set(used_ids)))
            #Just append the id to the normed name
            info['normed_name'] += f"-{info['id']}"

    return species_info


def get_parameter_data(xmltree: XMLLike,
                       schema_dict: fleur_schema.InputSchemaDict | fleur_schema.OutputSchemaDict,
                       inpgen_ready: bool = True,
                       write_ids: bool = True,
                       extract_econfig: bool = False,
                       allow_special_los: bool = True,
                       logger: Logger | None = None) -> dict[str, Any]:
    """
    This routine returns an python dictionary produced from the inp.xml
    file, which contains all the parameters needed to setup a new inp.xml from a inpgen
    input file to produce the same input (for parameters that the inpgen can control)

    :param xmltree: etree representing the fleur xml file
    :param schema_dict: schema dictionary corresponding to the file version
                        of the xmltree
    :param inpgen_ready: Bool, return a dict which can be inputted into inpgen while setting atoms
    :param write_ids: Bool, if True the atom ids are added to the atom namelists
    :param logger: logger object for logging warnings, errors

    :returns: dict, which will lead to the same inp.xml (in case if other defaults,
              which can not be controlled by input for inpgen, were changed)

    """
    from masci_tools.util.xml.converters import convert_fleur_lo, convert_fleur_electronconfig
    from masci_tools.io.common_functions import filter_out_empty_dict_entries

    # TODO: convert econfig
    # TODO: parse kpoints, somehow count is bad (if symmetry changes), mesh is not known, path cannot be specified

    ########
    parameters = {}

    # Create the cards

    # &input # most things are not needed for AiiDA here. or we ignor them for now.
    # film is set by the plugin depended on the structure
    # symor per default = False? to avoid input which fleur can't take

    with FleurXMLContext(xmltree, schema_dict, logger=logger) as root:

        # &comp namelist
        comp_dict = {}
        comp_dict['jspins'] = root.attribute('jspins')
        comp_dict['frcor'] = root.attribute('frcor', optional=True)
        comp_dict['ctail'] = root.attribute('ctail')
        comp_dict['kcrel'] = root.attribute('kcrel', optional=True)
        comp_dict['gmax'] = root.attribute('gmax')
        comp_dict['gmaxxc'] = root.attribute('gmaxxc')
        comp_dict['kmax'] = root.attribute('kmax')
        parameters['comp'] = filter_out_empty_dict_entries(comp_dict)

        species_info = _get_species_info(xmltree, schema_dict, logger=logger)
        for index, species in enumerate(root.iter('species')):

            atom_dict = {}
            atomlist_name = f'atom{index}'

            atom_z = species.attribute('atomicNumber')
            atom_name = species.attribute('name')
            if not inpgen_ready:
                atom_dict['z'] = atom_z
            atom_id = f"{atom_z}.{species_info[atom_name]['id']}"
            if write_ids:
                atom_dict['id'] = atom_id

            if schema_dict.inp_version <= (0, 31):
                atom_dict['ncst'] = species.attribute('coreStates')
            atom_dict['rmt'] = species.attribute('radius')
            atom_dict['dx'] = species.attribute('logIncrement')
            atom_dict['jri'] = species.attribute('gridPoints')
            atom_dict['lmax'] = species.attribute('lmax')
            atom_dict['lnonsph'] = species.attribute('lnonsphr')
            atom_dict['bmu'] = species.attribute('magMom', optional=True)

            atom_dict['element'] = species.attribute('element')

            if extract_econfig:
                if inpgen_ready:
                    atom_econfig = species.simple_xpath('electronConfig', list_return=True)
                    if len(atom_econfig) != 0:
                        atom_dict['econfig'] = convert_fleur_electronconfig(atom_econfig[0])
                else:
                    atom_dict['econfig'] = species.all_attributes('electronConfig', subtags=True, ignore={'flipSpins'})

            atom_lo = species.simple_xpath('lo', list_return=True)
            if len(atom_lo) != 0:
                atom_dict['lo'] = convert_fleur_lo(atom_lo, allow_special_los=allow_special_los)  #type:ignore[arg-type]

            parameters[atomlist_name] = filter_out_empty_dict_entries(atom_dict)

        # &soc namelist
        soc = root.attribute('l_soc', default=False)
        theta = root.attribute('theta', contains='soc', default=0.0)
        phi = root.attribute('phi', contains='soc', default=0.0)
        if soc:
            parameters['soc'] = {'theta': theta, 'phi': phi}

        # &kpt namelist
        if schema_dict.inp_version > (0, 31):
            list_name = root.attribute('listName')
            all_names = set(
                root.attribute('name', tag_name='kPointList', contains='kPointLists', list_return=True, optional=True))
            if list_name not in all_names:
                raise ValueError(f'Selected Kpoint list with the name: {list_name} does not exist\n'
                                 f'Available list names: {all_names}')

            with root.find('kPointList', contains='kPointLists', filters={'kPointList': {
                    'name': list_name
            }}) as kpoints:

                kpoint_type = kpoints.attribute('type', optional=True)
                if kpoint_type == 'mesh':
                    nx = kpoints.attribute('nx', optional=True)
                    ny = kpoints.attribute('ny', optional=True)
                    nz = kpoints.attribute('nz', optional=True)
                    if all(n is not None for n in (nx, ny, nz)):
                        parameters['kpt'] = {'div1': nx, 'div2': ny, 'div3': nz}

        # title
        title = root.text('comment', optional=True)
        if title:
            parameters['title'] = title.replace('\n', '').strip()

        # &exco
        exco_dict = {}
        exco_dict['xctyp'] = root.attribute(
            'name',
            tag_name='xcFunctional',
        )
        parameters['exco'] = filter_out_empty_dict_entries(exco_dict)
    # &film
    # TODO

    # &qss
    # TODO

    # lattice, not supported?

    return parameters


def get_structure_data(xmltree: XMLLike,
                       schema_dict: fleur_schema.InputSchemaDict | fleur_schema.OutputSchemaDict,
                       include_relaxations: bool = True,
                       convert_to_angstroem: bool = True,
                       normalize_kind_name: bool = True,
                       logger: Logger | None = None,
                       **kwargs: Any) -> tuple[list[AtomSiteProperties], np.ndarray, tuple[bool, bool, bool]]:
    """
    Get the structure defined in the given fleur xml file.

    .. warning::
        Only the explicit definition of the Bravais matrix is supported.
        Old inputs containing the `latnam` definitions are not supported

    .. warning::
        In versions ``0.5.0`` or later the output of the atom sites was restructured
        to be more interoperable with other IO functions (e.g. :py:func:`~masci_tools.io.fleur_inpgen.write_inpgen_file()`)
        The new format returns a list of :py:class:`~masci_tools.io.common_functions.AtomSiteProperties`
        instead of the list of tuples (position, symbol)

        For better compatibility this output is not default in ``0.5.0`` but instead
        is enabled by ``site_namedtuple=True`` and a DeprecationWarning is given when
        this argument is ``False``.

    .. note::
        In versions ``0.5.0`` or later the returned atom positions correspond to the relaxed
        structure if a ``relaxation`` section is present in the xmltree


    :param xmltree: etree representing the fleur xml file
    :param schema_dict: schema dictionary corresponding to the file version
                        of the xmltree
    :param include_relaxations: bool if True and a relaxation section is included
                                the resulting positions correspond to the relaxed structure
    :param logger: logger object for logging warnings, errors
    :param convert_to_angstroem: bool if True the bravais matrix is converted to angstroem

    :returns: tuple containing the structure information

    The tuple contains the following entries:

        1. :atom_data: list of (named)tuples containing the absolute positions and symbols of the atoms
        2. :cell: numpy array, bravais matrix of the given system
        3. :pbc: list of booleans, determines in which directions periodic boundary conditions are applicable

    .. versionchanged:: 0.7.0
        The default for `site_namedtuple` is set to `True`

    .. versionchanged:: 0.10.0
        The argument `site_namedtuple` was deprecated. The old output is no longer supported. If the
        argument `site_namedtuple` is passed a deprecation warning is shown

    """
    from masci_tools.io.common_functions import rel_to_abs, rel_to_abs_f, abs_to_rel, abs_to_rel_f
    from masci_tools.io.common_functions import find_symmetry_relation
    from masci_tools.util.constants import BOHR_A

    if 'site_namedtuple' in kwargs:
        warnings.warn(
            'The argument site_namedtuple is deprecated and has no effect.'
            'The output is always given in AtomSiteProperties', DeprecationWarning)

    cell, pbc = get_cell(xmltree, schema_dict, logger=logger, convert_to_angstroem=convert_to_angstroem)
    species_info = _get_species_info(xmltree, schema_dict, logger=None)

    atom_data: list[AtomSiteProperties] = []
    with FleurXMLContext(xmltree, schema_dict, logger=logger) as root:

        #Read relaxation information if available
        displacements = None
        rotations, shifts = None, None
        if include_relaxations and schema_dict.inp_version >= (0, 29):
            if root.tag_exists('relaxation'):
                relax_info = get_relaxation_information(xmltree, schema_dict, logger=logger)
                #We still read in the normal atom positions since the displacements are provided
                #per atomtype
                displacements = relax_info['displacements']
                if convert_to_angstroem:
                    displacements = [np.array(displace) * BOHR_A for displace in displacements]
                rotations, shifts = get_symmetry_information(xmltree, schema_dict, logger=logger)

                if len(displacements) != root.number_nodes('atomGroup'):
                    raise ValueError(
                        f"Did not get the right number of relaxed positions. Expected {root.number_nodes('atomGroup')} got {len(displacements)}"
                    )

        for index, group in enumerate(root.iter('atomGroup')):

            atom_positions: list[list[float]] = []

            absolute_positions = group.text('absPos', list_return=True, optional=True)
            relative_positions = group.text('relPos', list_return=True, optional=True)
            film_positions = group.text('filmPos', list_return=True, optional=True)

            if convert_to_angstroem:
                atom_positions = [list(np.array(pos) * BOHR_A) for pos in absolute_positions]
            else:
                atom_positions = absolute_positions

            atom_positions.extend(rel_to_abs(pos, cell) for pos in relative_positions)

            for film_pos in film_positions:
                film_pos = rel_to_abs_f(film_pos, cell)
                if convert_to_angstroem:
                    film_pos[2] *= BOHR_A
                atom_positions.append(film_pos)

            if len(atom_positions) == 0:
                raise ValueError('Failed to read atom positions for group')

            if displacements:
                representative_pos = np.array(atom_positions[0])

                if rotations is None or shifts is None:
                    raise ValueError('Symmetry information is required but not available')

                if len(film_positions) != 0:
                    rel_displace = abs_to_rel_f(displacements[index], cell, pbc)
                    rel_representative_pos = abs_to_rel_f(representative_pos, cell, pbc)
                    rel_displace[2] = rel_displace[2] / cell[2, 2]
                    rel_representative_pos[2] = rel_representative_pos[2] / cell[2, 2]
                else:
                    rel_displace = abs_to_rel(displacements[index], cell)
                    rel_representative_pos = abs_to_rel(representative_pos, cell)

                for pos_indx, pos in enumerate(atom_positions):
                    rot, shift = find_symmetry_relation(representative_pos,
                                                        pos,
                                                        rotations,
                                                        shifts,
                                                        cell,
                                                        relative_pos=False,
                                                        film=len(film_positions) != 0)

                    #More explicit than it needs to be
                    #but analogous to fleur
                    rot_pos = np.matmul(rot, rel_representative_pos) + shift
                    site_displace = np.matmul(rot, rel_representative_pos + rel_displace) + shift
                    site_displace = site_displace - rot_pos

                    if len(film_positions) != 0:
                        site_displace = rel_to_abs_f(site_displace, cell)
                        site_displace[2] *= cell[2, 2]
                    else:
                        site_displace = rel_to_abs(site_displace, cell)

                    atom_positions[pos_indx] = list(np.array(atom_positions[pos_indx]) + np.array(site_displace))

            group_species = group.attribute('species')
            element = species_info[group_species]['element']
            if normalize_kind_name:
                normed_name = species_info[group_species]['normed_name']
                if normed_name != group_species:
                    if logger is None:
                        warnings.warn(
                            f'Normalized species name {group_species} to {normed_name}. '
                            "Use the option 'normed_kind_name=False' to preserve the original species name",
                            UserWarning)
                    else:
                        logger.warning(f'Normalized species name {group_species} to {normed_name}. '
                                       "Use the option 'normed_kind_name=False' to preserve the original species name")
                    group_species = normed_name

            atom_data.extend(
                AtomSiteProperties(position=pos, symbol=element, kind=group_species) for pos in atom_positions)

    return atom_data, cell, pbc


@schema_dict_version_dispatch(output_schema=False)
def get_kpoints_data(
    xmltree: XMLLike,
    schema_dict: fleur_schema.InputSchemaDict | fleur_schema.OutputSchemaDict,
    name: str | None = None,
    index: int | None = None,
    only_used: bool = False,
    logger: Logger | None = None,
    convert_to_angstroem: bool = True
) -> tuple[list[list[float]] | dict[str, list[list[float]]], list[float] | dict[str, list[float]], np.ndarray, tuple[
        bool, bool, bool]]:
    """
    Get the kpoint sets defined in the given fleur xml file.

    .. warning::
        For file versions before Max5 arguments `name`, `index` and `only_used`
        have no effect

    :param xmltree: etree representing the fleur xml file
    :param schema_dict: schema dictionary corresponding to the file version
                        of the xmltree
    :param name: str, optional, if given only the kpoint set with the given name
                 is returned
    :param index: int, optional, if given only the kpoint set with the given index
                  is returned
    :param only_used: bool if True only the kpoint list used in the calculation is returned
    :param logger: logger object for logging warnings, errors
    :param convert_to_angstroem: bool if True the bravais matrix is converted to angstroem

    :returns: tuple containing the kpoint information

    The tuple contains the following entries:

        1. :kpoints: dict or list (list if there is only one kpoint set),
                     containing the coordinates of the kpoints
        2. :weights: dict or list (list if there is only one kpoint set),
                     containing the weights of the kpoints
        3. :cell: numpy array, bravais matrix of the given system
        4. :pbc: list of booleans, determines in which directions periodic boundary conditions are applicable

    """

    if name is not None and index is not None:
        raise ValueError('Only provide one of index or name to select kpoint lists')

    if only_used and (name is not None or index is not None):
        raise ValueError('Either use only_used=False and provide the name/index or use only_used=True. Not both')

    cell, pbc = get_cell(xmltree, schema_dict, logger=logger, convert_to_angstroem=convert_to_angstroem)

    with FleurXMLContext(xmltree, schema_dict, logger=logger) as root:

        if only_used:
            name = root.attribute('listName')

        filters: FilterType = {}
        if name is not None:
            filters['kPointList'] = {'name': name}
        if index is not None:
            filters['kPointList'] = {'index': index + 1 if index >= 0 else index}

        if root.number_nodes('kPointList', contains='kPointLists', filters=filters) == 0:
            labels = root.attribute('name', tag_name='kPointList', contains='kPointLists', list_return=True)
            if only_used:
                raise ValueError(f'Found no Kpoint list with the name: {name}'
                                 f'Available list names: {labels}'
                                 'The listName attribute is not consistent with the rest of the input')
            if name is not None:
                raise ValueError(f'Found no Kpoint list with the name: {name}'
                                 f'Available list names: {labels}')
            raise ValueError('No Kpoint lists found in the given inp.xml, matching the criteria could be found'
                             f'Available list names: {labels}')

        kpoints_data = {}
        weights_data = {}
        for kpointlist in root.iter('kPointList', contains='kPointLists', filters=filters):

            label = kpointlist.attribute('name')
            kpoints = kpointlist.text('kPoint', list_return=True)
            weights = kpointlist.attribute('weight', list_return=True)

            if not isinstance(kpoints[0], list):
                kpoints = [kpoints]
                weights = [weights]

            kpoints_data[label] = kpoints
            weights_data[label] = weights

    if len(kpoints_data) == 1:
        _, kpoints_data = kpoints_data.popitem()
        _, weights_data = weights_data.popitem()

    return kpoints_data, weights_data, cell, pbc


@get_kpoints_data.register(max_version='0.31')
def get_kpoints_data_max4(
        xmltree: XMLLike,
        schema_dict: fleur_schema.InputSchemaDict | fleur_schema.OutputSchemaDict,
        name: str | None = None,
        index: int | None = None,
        only_used: bool = False,
        logger: Logger | None = None,
        convert_to_angstroem: bool = True
) -> tuple[list[list[float]], list[float], np.ndarray, tuple[bool, bool, bool]]:
    """
    Get the kpoint sets defined in the given fleur xml file.

    .. note::
        This function is specific to file version before and including the
        Max4 release of fleur

    :param xmltree: etree representing the fleur xml file
    :param schema_dict: schema dictionary corresponding to the file version
                        of the xmltree
    :param logger: logger object for logging warnings, errors
    :param convert_to_angstroem: bool if True the bravais matrix is converted to angstroem
    :param only_used: (Has no effect for Max4) bool if True only the kpoint list used in the calculation is returned
    :param name: (Has no effect for Max4)
    :param index: (Has no effect for Max4)

    :returns: tuple containing the kpoint information

    The tuple contains the following entries:

        1. :kpoints: list containing the coordinates of the kpoints
        2. :weights: list containing the weights of the kpoints
        3. :cell: numpy array, bravais matrix of the given system
        4. :pbc: list of booleans, determines in which directions periodic boundary conditions are applicable

    """
    cell, pbc = get_cell(xmltree, schema_dict, logger=logger, convert_to_angstroem=convert_to_angstroem)

    with FleurXMLContext(xmltree, schema_dict, logger=logger) as root:

        kpointlist_tag = root.simple_xpath('kPointList',
                                           list_return=True,
                                           not_contains=['altKPoint', 'numericalParameters'])

        if len(kpointlist_tag) == 0:
            if getattr(schema_dict, 'out_version', None) is not None:
                kpointlist_tag = root.simple_xpath('kPointList', list_return=True, contains='numericalParameters')
            else:
                raise ValueError('No Kpoint lists found in the given inp.xml')

        with root.nested(kpointlist_tag[0]) as kpointlist:
            kpoints = kpointlist.text('kPoint', list_return=True)
            weights = kpointlist.attribute('weight', list_return=True)

    return kpoints, weights, cell, pbc


@schema_dict_version_dispatch(output_schema=False)
def get_special_kpoints(
    xmltree: XMLLike,
    schema_dict: fleur_schema.InputSchemaDict | fleur_schema.OutputSchemaDict,
    name: str | None = None,
    index: int | None = None,
    only_used: bool = False,
    logger: Logger | None = None,
) -> list[tuple[int, str]] | dict[str, list[tuple[int, str]]]:
    """
    Extract the labeled special kpoints from the given kpointlist

    .. warning::
        Only implemented for versions starting with Max5

    :param xmltree: etree representing the fleur xml file
    :param schema_dict: schema dictionary corresponding to the file version
                        of the xmltree
    :param name: str, optional, if given only the kpoint set with the given name
                 is returned
    :param index: int, optional, if given only the kpoint set with the given index
                  is returned
    :param only_used: bool if True only the kpoint list used in the calculation is returned
    :param logger: logger object for logging warnings, errors

    :returns: list of tuples (index, label) for multiple kpoint sets a dict with the names containing
              the list of tuples is returned
    """

    if name is not None and index is not None:
        raise ValueError('Only provide one of index or name to select kpoint lists')

    if only_used and (name is not None or index is not None):
        raise ValueError('Either use only_used=False and provide the name/index or use only_used=True. Not both')

    with FleurXMLContext(xmltree, schema_dict, logger=logger) as root:

        if only_used:
            name = root.attribute('listName')

        filters: FilterType = {}
        if name is not None:
            filters['kPointList'] = {'name': name}
        if index is not None:
            filters['kPointList'] = {'index': index + 1 if index >= 0 else index}

        if root.number_nodes('kPointList', contains='kPointLists', filters=filters) == 0:
            labels = root.attribute('name', tag_name='kPointList', contains='kPointLists', list_return=True)
            if only_used:
                raise ValueError(f'Found no Kpoint list with the name: {name}'
                                 f'Available list names: {labels}'
                                 'The listName attribute is not consistent with the rest of the input')
            if name is not None:
                raise ValueError(f'Found no Kpoint list with the name: {name}'
                                 f'Available list names: {labels}')
            raise ValueError('No Kpoint lists found in the given inp.xml, matching the criteria could be found'
                             f'Available list names: {labels}')

        special_kpoints = {}
        for kpointlist in root.iter('kPointList', contains='kPointLists', filters=filters):

            label = kpointlist.attribute('name')

            labelled_points = kpointlist.simple_xpath('kPoint',
                                                      filters={'kPoint': {
                                                          'label': {
                                                              '!=': ''
                                                          }
                                                      }},
                                                      list_return=True)

            #yapf: disable
            special_kpoints[label] = [
                (kpointlist.node.index(kpoint), str(kpoint.attrib['label'])) for kpoint in labelled_points  #type:ignore[union-attr]
            ]
            #yapf: enable

    if len(special_kpoints) == 1:
        _, special_kpoints = special_kpoints.popitem()  #type:ignore[assignment]

    return special_kpoints


@get_special_kpoints.register(max_version='0.31')
def get_special_kpoints_max4(
    xmltree: XMLLike,
    schema_dict: fleur_schema.InputSchemaDict | fleur_schema.OutputSchemaDict,
    name: str | None = None,
    index: int | None = None,
    only_used: bool = False,
    logger: Logger | None = None,
) -> list[tuple[int, str]] | dict[str, list[tuple[int, str]]]:
    """
    Extract the labeled special kpoints from the given kpointlist

    .. warning::
        Only implemented for versions starting with Max5

    :param xmltree: etree representing the fleur xml file
    :param schema_dict: schema dictionary corresponding to the file version
                        of the xmltree
    :param name: str, optional, if given only the kpoint set with the given name
                 is returned
    :param index: int, optional, if given only the kpoint set with the given index
                  is returned
    :param only_used: bool if True only the kpoint list used in the calculation is returned
    :param logger: logger object for logging warnings, errors

    :returns: list of tuples (index, label) for multiple kpoint sets a dict with the names containing
              the list of tuples is returned
    """

    raise NotImplementedError(
        f"'get_special_kpoints' is not implemented for inputs of version '{schema_dict['inp_version']}'")


@schema_dict_version_dispatch(output_schema=False)
def get_relaxation_information(xmltree: XMLLike,
                               schema_dict: fleur_schema.InputSchemaDict | fleur_schema.OutputSchemaDict,
                               logger: Logger | None = None) -> dict[str, Any]:
    """
    Get the relaxation information from the given fleur XML file. This includes the current
    displacements, energy and posforce evolution

    :param xmltree: etree representing the fleur xml file
    :param schema_dict: schema dictionary corresponding to the file version
                        of the xmltree
    :param logger: logger object for logging warnings, errors

    :returns: dict with the relaxation information

    :raises ValueError: If no relaxation section is included in the xml tree
    """

    with FleurXMLContext(xmltree, schema_dict, logger=logger) as root:

        if not root.tag_exists('relaxation'):
            raise ValueError('No relaxation information included in the given XML file')

        out_dict = {}
        with root.find('relaxation') as relax_tag:

            out_dict['displacements'] = relax_tag.text('displace', list_return=True)
            out_dict['energies'] = relax_tag.attribute('energy', list_return=True)
            out_dict['posforces'] = [step.text('posforce', list_return=True) for step in relax_tag.iter('step')]

    return out_dict


@get_relaxation_information.register(max_version='0.28')
def get_relaxation_information_pre029(xmltree: XMLLike,
                                      schema_dict: fleur_schema.InputSchemaDict | fleur_schema.OutputSchemaDict,
                                      logger: Logger | None = None) -> dict[str, Any]:
    """
    Get the relaxation information from the given fleur XML file. This includes the current
    displacements, energy and posforce evolution

    :param xmltree: etree representing the fleur xml file
    :param schema_dict: schema dictionary corresponding to the file version
                        of the xmltree
    :param logger: logger object for logging warnings, errors

    :returns: dict with the relaxation information

    :raises ValueError: If no relaxation section is included in the xml tree
    """
    raise NotImplementedError(
        f"'get_relaxation_information' is not implemented for inputs of version '{schema_dict['inp_version']}'")


def get_symmetry_information(xmltree: XMLLike,
                             schema_dict: fleur_schema.InputSchemaDict | fleur_schema.OutputSchemaDict,
                             logger: Logger | None = None) -> tuple[list[np.ndarray], list[np.ndarray]]:
    """
    Get the symmetry information from the given fleur XML file. This includes the
    rotation matrices and shifts defined in the ``symmetryOperations`` tag.

    .. note::
        Only the explicit definition of the used symmetry operations in the xml file
        is supported.

    :param xmltree: etree representing the fleur xml file
    :param schema_dict: schema dictionary corresponding to the file version
                        of the xmltree
    :param logger: logger object for logging warnings, errors

    :returns: tuple of the rotations and their respective shifts

    :raises ValueError: If no symmetryOperations section is included in the xml tree
    """

    with FleurXMLContext(xmltree, schema_dict, logger=logger) as root:

        if not root.tag_exists('symmetryOperations'):
            raise ValueError('No explicit symmetry information included in the given XML file')

        rotations = []
        shifts = []
        for symop in root.iter('symOp'):
            row1 = symop.text('row-1')
            row2 = symop.text('row-2')
            row3 = symop.text('row-3')

            rot = np.array([row1[:3], row2[:3], row3[:3]]).astype(int)
            shift = np.array([row1[3], row2[3], row3[3]])

            rotations.append(rot)
            shifts.append(shift)

    return rotations, shifts



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module provides easy functions for loading a input/output xml file of
fleur and providing a parsed xml etree together with its corresponding schema dict
"""
from __future__ import annotations
import logging

from lxml import etree
import warnings
import os
from pathlib import Path
from functools import partial
from logging import Logger
from typing import Callable, Any, Generator
from contextlib import _GeneratorContextManager, contextmanager

from masci_tools.io.parsers import fleur_schema
from masci_tools.util.typing import XMLFileLike, XMLLike
from masci_tools.util.xml.common_functions import normalize_xmllike

__all__ = ('load_inpxml', 'load_outxml', 'FleurXMLContext', 'get_constants', 'load_outxml_and_check_for_broken_xml',
           '_EvalContext')


def load_inpxml(inpxmlfile: XMLFileLike,
                logger: Logger | None = None,
                base_url: str | Path | None = None,
                **kwargs: Any) -> tuple[etree._ElementTree, fleur_schema.InputSchemaDict]:
    """
    Loads a inp.xml file for fleur together with its corresponding schema dictionary

    :param inpxmlfile: either path to the inp.xml file, opened file handle (in bytes modes i.e. rb)
                       or a xml etree to be parsed

    :returns: parsed xmltree of the inpxmlfile and the schema dictionary
              for the corresponding input version
    """
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.io.parsers.fleur_schema import InputSchemaDict

    xml_parse_func: Callable = etree.parse
    if isinstance(inpxmlfile, (str, bytes, Path)) and not os.path.isfile(inpxmlfile):
        xml_parse_func = etree.fromstring
        if base_url is None:
            warnings.warn('You provided a string of content but no base_url argument.'
                          'Setting it to the current working directory.'
                          'If the tree contains xinclude tags these could fail')
            base_url = os.getcwd()
        elif isinstance(base_url, Path):
            base_url = os.fspath(base_url)
        xml_parse_func = partial(xml_parse_func, base_url=base_url)

    if isinstance(inpxmlfile, etree._ElementTree):
        xmltree = inpxmlfile
    else:
        parser = etree.XMLParser(attribute_defaults=True, encoding='utf-8', **kwargs)

        try:
            xmltree = xml_parse_func(inpxmlfile, parser)
        except etree.XMLSyntaxError as msg:
            if logger is not None:
                logger.exception('Failed to parse input file')
            raise ValueError(f'Failed to parse input file: {msg}') from msg

    if etree.iselement(xmltree):
        xmltree = xmltree.getroottree()

    if xmltree is None:
        if logger is not None:
            logger.error('No XML tree generated. Check that the given file exists')
        raise ValueError('No XML tree generated. Check that the given file exists')

    version = eval_xpath(xmltree, '//@fleurInputVersion')
    if not version:
        if logger is not None:
            logger.error('Failed to extract inputVersion')
        raise ValueError('Failed to extract inputVersion')
    version = str(version)

    if logger is not None:
        logger.info('Got Fleur input file with file version %s', version)

    schema_dict = InputSchemaDict.fromVersion(version, logger=logger)

    return xmltree, schema_dict


def load_outxml(outxmlfile: XMLFileLike,
                logger: Logger | None = None,
                base_url: str | Path | None = None,
                **kwargs: Any) -> tuple[etree._ElementTree, fleur_schema.OutputSchemaDict]:
    """
    Loads a out.xml file for fleur together with its corresponding schema dictionary.
    Also returns whether the XML file had to be parsed with `recover=True`

    :param outxmlfile: either path to the out.xml file, opened file handle (in bytes modes i.e. rb)
                       or a xml etree to be parsed

    :returns: parsed xmltree of the outxmlfile and the schema dictionary
              for the corresponding output version
    """
    xmltree, schema_dict, _ = load_outxml_and_check_for_broken_xml(outxmlfile, logger, base_url, **kwargs)
    return xmltree, schema_dict


def load_outxml_and_check_for_broken_xml(
        outxmlfile: XMLFileLike,
        logger: Logger | None = None,
        base_url: str | Path | None = None,
        **kwargs: Any) -> tuple[etree._ElementTree, fleur_schema.OutputSchemaDict, bool]:
    """
    Loads a out.xml file for fleur together with its corresponding schema dictionary.
    Also returns whether the XML file had to be parsed with `recover=True`

    :param outxmlfile: either path to the out.xml file, opened file handle (in bytes modes i.e. rb)
                       or a xml etree to be parsed

    :returns: parsed xmltree of the outxmlfile and the schema dictionary
              for the corresponding output version and bool indicating whether the outxml is broken
    """
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.io.parsers.fleur_schema import OutputSchemaDict

    xml_parse_func: Callable = etree.parse
    if isinstance(outxmlfile, (str, bytes, Path)) and not os.path.isfile(outxmlfile):
        xml_parse_func = etree.fromstring
        if base_url is None:
            warnings.warn('You provided a string of content but no base_url argument.'
                          'Setting it to the current working directory.'
                          'If the tree contains xinclude tags these could fail')
            base_url = os.getcwd()
        elif isinstance(base_url, Path):
            base_url = os.fspath(base_url)
        xml_parse_func = partial(xml_parse_func, base_url=base_url)

    outfile_broken = False

    if isinstance(outxmlfile, etree._ElementTree):
        xmltree = outxmlfile
    else:
        parser = etree.XMLParser(attribute_defaults=True, recover=False, encoding='utf-8', **kwargs)

        try:
            xmltree = xml_parse_func(outxmlfile, parser)
        except etree.XMLSyntaxError:
            outfile_broken = True
            if logger is None:
                warnings.warn('The out.xml file is broken I try to repair it.')
            else:
                logger.warning('The out.xml file is broken I try to repair it.')

        if outfile_broken:
            # repair xmlfile and try to parse what is possible.
            parser = etree.XMLParser(attribute_defaults=True, recover=True, encoding='utf-8', **kwargs)

            try:
                xmltree = xml_parse_func(outxmlfile, parser)
            except etree.XMLSyntaxError as err:
                raise ValueError('Skipping the parsing of the XML file. Repairing was not possible.') from err

    if etree.iselement(xmltree):
        xmltree = xmltree.getroottree()

    if xmltree is None:
        if logger is not None:
            logger.error('No XML tree generated. Check that the given file exists')
        raise ValueError('No XML tree generated. Check that the given file exists')

    out_version = eval_xpath(xmltree, '//@fleurOutputVersion')
    if not out_version:
        if logger is not None:
            logger.error('Failed to extract outputVersion')
        raise ValueError('Failed to extract outputVersion')
    out_version = str(out_version)

    if out_version == '0.27':
        program_version = str(eval_xpath(xmltree, '//programVersion/@version'))
        if program_version == 'fleur 32':
            #Max5 release (before bugfix)
            out_version = '0.33'
            inp_version = '0.33'
            if logger is not None:
                logger.warning("Ignoring '0.27' outputVersion for MaX5.0 release")
            else:
                warnings.warn("Ignoring '0.27' outputVersion for MaX5.0 release")
        elif program_version == 'fleur 31':
            #Max4 release
            out_version = '0.31'
            inp_version = '0.31'
            if logger is not None:
                logger.warning("Ignoring '0.27' outputVersion for MaX4.0 release")
            else:
                warnings.warn("Ignoring '0.27' outputVersion for MaX4.0 release")
        elif program_version == 'fleur 30':
            #Max3.1 release
            out_version = '0.30'
            inp_version = '0.30'
            if logger is not None:
                logger.warning("Ignoring '0.27' outputVersion for MaX3.1 release")
            else:
                warnings.warn("Ignoring '0.27' outputVersion for MaX3.1 release")
        elif program_version == 'fleur 27':
            #Max3.1 release
            out_version = '0.29'
            inp_version = '0.29'
            if logger is not None:
                logger.warning("Found version before MaX3.1 release falling back to file version '0.29'")
            warnings.warn(
                'out.xml files before the MaX3.1 release are not explicitly supported.'
                ' No guarantee is given that the parser will work without error', UserWarning)
        else:
            if logger is not None:
                logger.error("Unknown fleur version: File-version '%s' Program-version '%s'", out_version,
                             program_version)
            raise ValueError(f"Unknown fleur version: File-version '{out_version}' Program-version '{program_version}'")
    else:
        inp_version: str = eval_xpath(xmltree, '//@fleurInputVersion')  #type:ignore
        if not inp_version:
            raise ValueError('Failed to extract inputVersion')
        inp_version = str(inp_version)

    schema_dict = OutputSchemaDict.fromVersion(out_version, inp_version=inp_version, logger=logger)

    return xmltree, schema_dict, outfile_broken


class _EvalContext:
    """
    Contextmanager to hold the relevant datastructures for evaluating values from XML files

    This holds:

        :param etree_or_element: The XML tree of the file
        :param schema_dict: The corresponding SchemaDict
        :param constants: The dictionary containing the defined mathematical constants
        :param logger: The configured logger instance
    """

    def __init__(self,
                 etree_or_element: XMLLike | etree.XPathElementEvaluator,
                 schema_dict: fleur_schema.InputSchemaDict | fleur_schema.OutputSchemaDict,
                 constants: dict[str, float] | None = None,
                 logger: logging.Logger | None = None) -> None:

        self.node: etree._Element | etree.XPathElementEvaluator
        if not isinstance(etree_or_element, etree.XPathElementEvaluator):
            self.node = normalize_xmllike(etree_or_element)
        else:
            self.node = etree_or_element
        self.schema_dict = schema_dict
        self.logger = logger
        self.constants = constants or get_constants(self.node, self.schema_dict, self.logger)

    def attribute(self, name: str, default: Any | None = None, **kwargs: Any) -> Any:
        """
        Alias for :py:func:`~masci_tools.util.schema_dict_util.evaluate_attribute()`

        :param name: str, name of the attribute
        :param default: value to return, if there are no values found
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param iteration_path: bool if True and the SchemaDict is of an output schema an absolute path into
                            the iteration element is constructed
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details

        Kwargs:
            :param tag_name: str, name of the tag where the attribute should be parsed
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
            :param exclude: list of str, here specific types of attributes can be excluded
                            valid values are: settable, settable_contains, other
            :param list_return: if True, the returned quantity is always a list even if only one element is in it
            :param optional: bool, if True and no logger given none or an empty list is returned

        :returns: list or single value, converted in convert_xml_attribute
        """
        from masci_tools.util.schema_dict_util import evaluate_attribute
        if default is not None:
            kwargs['optional'] = True
        res = evaluate_attribute(self.node,
                                 self.schema_dict,
                                 name,
                                 logger=self.logger,
                                 constants=self.constants,
                                 **kwargs)
        if res is None and default is not None:
            return default
        return res

    def text(self, name: str, **kwargs: Any) -> Any:
        """
        Alias for :py:func:`~masci_tools.util.schema_dict_util.evaluate_text()`

        :param name: str, name of the tag
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param iteration_path: bool if True and the SchemaDict is of an output schema an absolute path into
                            the iteration element is constructed
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details

        Kwargs:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
            :param list_return: if True, the returned quantity is always a list even if only one element is in it
            :param optional: bool, if True and no logger given none or an empty list is returned

        :returns: list or single value, converted in convert_xml_text
        """
        from masci_tools.util.schema_dict_util import evaluate_text
        return evaluate_text(self.node, self.schema_dict, name, logger=self.logger, constants=self.constants, **kwargs)

    def all_attributes(self, name: str, **kwargs: Any) -> Any:
        """
        Alias for :py:func:`~masci_tools.util.schema_dict_util.evaluate_tag()`

        :param name: str, name of the tag
        :param subtags: optional bool, if True the subtags of the given tag are evaluated
        :param text: optional bool, if True the text of the tag is also parsed
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param iteration_path: bool if True and the SchemaDict is of an output schema an absolute path into
                            the iteration element is constructed
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details

        Kwargs:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
            :param only_required: bool (optional, default False), if True only required attributes are parsed
            :param ignore: list of str (optional), attributes not to parse
            :param list_return: if True, the returned quantity is always a list even if only one element is in it
            :param strict_missing_error: if True, and no logger is given an error is raised if any attribute is not found

        """
        from masci_tools.util.schema_dict_util import evaluate_tag
        return evaluate_tag(self.node, self.schema_dict, name, logger=self.logger, constants=self.constants, **kwargs)

    def parent_attributes(self, name: str, **kwargs: Any) -> Any:
        """
        Alias for :py:func:`~masci_tools.util.schema_dict_util.evaluate_parent_tag()`

        :param name: str, name of the tag
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param iteration_path: bool if True and the SchemaDict is of an output schema an absolute path into
                            the iteration element is constructed
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details

        Kwargs:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
            :param only_required: bool (optional, default False), if True only required attributes are parsed
            :param ignore: list of str (optional), attributes not to parse
            :param list_return: if True, the returned quantity is always a list even if only one element is in it
            :param strict_missing_error: if True, and no logger is given an error is raised if any attribute is not found

        :returns: dict, with attribute values converted via convert_xml_attribute
        """
        from masci_tools.util.schema_dict_util import evaluate_parent_tag
        return evaluate_parent_tag(self.node,
                                   self.schema_dict,
                                   name,
                                   logger=self.logger,
                                   constants=self.constants,
                                   **kwargs)

    def single_value(self, name: str, **kwargs: Any) -> Any:
        """
        Alias for :py:func:`~masci_tools.util.schema_dict_util.evaluate_single_value_tag()`

        :param name: str, name of the tag
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation

        Kwargs:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
            :param only_required: bool (optional, default False), if True only required attributes are parsed
            :param ignore: list of str (optional), attributes not to parse
            :param list_return: if True, the returned quantity is always a list even if only one element is in it
            :param strict_missing_error: if True, and no logger is given an error is raised if any attribute is not found
            :param iteration_path: bool if True and the SchemaDict is of an output schema an absolute path into
                                the iteration element is constructed
            :param filters: Dict specifying constraints to apply on the xpath.
                            See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details

        :returns: value and unit, both converted in convert_xml_attribute
        """
        from masci_tools.util.schema_dict_util import evaluate_single_value_tag
        return evaluate_single_value_tag(self.node,
                                         self.schema_dict,
                                         name,
                                         logger=self.logger,
                                         constants=self.constants,
                                         **kwargs)

    def tag_exists(self, name: str, **kwargs: Any) -> bool:
        """
        Alias for :py:func:`~masci_tools.util.schema_dict_util.tag_exists()`

        :param name: str, name of the tag
        :param iteration_path: bool if True and the SchemaDict is of an output schema an absolute path into
                            the iteration element is constructed
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details

        Kwargs:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path

        :returns: bool, True if any tag exists
        """
        from masci_tools.util.schema_dict_util import tag_exists
        return tag_exists(self.node, self.schema_dict, name, logger=self.logger, **kwargs)

    def number_nodes(self, name: str, **kwargs: Any) -> int:
        """
        Alias of :py:func:`~masci_tools.util.schema_dict_util.get_number_of_nodes()`

        :param name: str, name of the tag

        Kwargs:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
            :param iteration_path: bool if True and the SchemaDict is of an output schema an absolute path into
                                   the iteration element is constructed
            :param filters: Dict specifying constraints to apply on the xpath.
                            See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details

        :returns: number of nodes of the given tag
        """
        from masci_tools.util.schema_dict_util import get_number_of_nodes
        return get_number_of_nodes(self.node, self.schema_dict, name, logger=self.logger, **kwargs)

    def attribute_exists(self, name: str, **kwargs: Any) -> bool:
        """
        Alias for :py:func:`~masci_tools.util.schema_dict_util.attrib_exists()`

        :param name: str, name of the attribute
        :param iteration_path: bool if True and the SchemaDict is of an output schema an absolute path into
                            the iteration element is constructed
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details

        Kwargs:
            :param tag_name: str, name of the tag where the attribute should be parsed
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
            :param exclude: list of str, here specific types of attributes can be excluded
                            valid values are: settable, settable_contains, other

        :returns: bool, True if any tag with the attribute exists
        """
        from masci_tools.util.schema_dict_util import attrib_exists
        return attrib_exists(self.node, self.schema_dict, name, logger=self.logger, **kwargs)

    def simple_xpath(self, name: str, **kwargs: Any) -> list[etree._Element] | etree._Element:
        """
        Alias for :py:func:`~masci_tools.util.schema_dict_util.eval_simple_xpath()`

        :param name: str, name of the tag
        :param iteration_path: bool if True and the SchemaDict is of an output schema an absolute path into
                            the iteration element is constructed
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details
        :param list_return: bool, if True a list is always returned

        Kwargs:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path

        :returns: etree Elements obtained via the simple xpath expression
        """
        from masci_tools.util.schema_dict_util import eval_simple_xpath
        return eval_simple_xpath(self.node, self.schema_dict, name, logger=self.logger, **kwargs)

    def find(self, name: str, **kwargs: Any) -> _GeneratorContextManager[_EvalContext]:  #pylint: disable=unsubscriptable-object
        """
        Finds the first element for the given name and constraints and gives a nested
        context for this element, i.e. inheriting the schema_dict, constants and logger

        Example::

            with FleurXMLContext(xmltree, schema_dict) as root:
                #Operations happen on the root here
                jspins = root.attribute('jspins')
                with root.find('species') as species:
                    #Operations in this block happen on the first species node
                    radius = species.attribute('radius')

        :param name: str, name of the tag
        :param iteration_path: bool if True and the SchemaDict is of an output schema an absolute path into
                            the iteration element is constructed
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details

        Kwargs:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
        """
        if 'list_return' in kwargs:
            raise ValueError('The argument list_return is not allowed in find()')

        nodes = self.simple_xpath(name, **kwargs, list_return=True)
        if nodes:
            return self.nested(nodes[0])
        raise ValueError(f'No nodes found for name {name}')

    @contextmanager
    def nested(self, etree_or_element: XMLLike) -> Generator[_EvalContext, None, None]:
        """
        Create a nested context from the current one
        inheriting the schema_dict, constants and logger only replacing the
        XML element

        Example of explicit usage::

            with FleurXMLContext(xmltree, schema_dict) as root:
                #Operations happen on the root here
                jspins = root.attribute('jspins')

                all_species_tag = root.simple_xpath('atomspecies')
                with root.nested(all_species_tag) as all_species:
                    #Operations happen on the 'atomSpecies' tag
                    mt_radii = all_species.attribute('radius')

        :param etree_or_element: Element to use for evaluation in the nested context
        """
        yield _EvalContext(etree_or_element, self.schema_dict, self.constants, logger=self.logger)

    def iter(self, name: str, **kwargs: Any) -> Generator[_EvalContext, None, None]:
        """
        Finds all elements for the given name and constraints and gives nested
        contexts for these elements to be iterated over,
        i.e. inheriting the schema_dict, constants and logger

        Example::

            with FleurXMLContext(xmltree, schema_dict) as root:
                #Operations happen on the root here
                jspins = root.attribute('jspins')
                for species in root.iter('species'):
                    #Operations happen on one species node in each iteration of the
                    #for loop. The order is the same as they appear in the XMl file
                    radius = species.attribute('radius')

        :param name: str, name of the tag
        :param iteration_path: bool if True and the SchemaDict is of an output schema an absolute path into
                            the iteration element is constructed
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details

        Kwargs:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
        """
        if 'list_return' in kwargs:
            raise ValueError('The argument list_return is not allowed in iter()')
        nodes = self.simple_xpath(name, **kwargs, list_return=True)
        for node in nodes:
            with self.nested(node) as ctx:
                yield ctx


@contextmanager
def FleurXMLContext(etree_or_element: XMLLike | etree.XPathElementEvaluator,
                    schema_dict: fleur_schema.InputSchemaDict | fleur_schema.OutputSchemaDict,
                    constants: dict[str, float] | None = None,
                    logger: logging.Logger | None = None) -> Generator[_EvalContext, None, None]:
    """
    Contextmanager to hold the relevant datastructures for evaluating values from XML files

    This holds:

        :param etree_or_element: The XML tree of the file
        :param schema_dict: The corresponding SchemaDict
        :param constants: The dictionary containing the defined mathematical constants
        :param logger: The configured logger instance

    The following methods are available:
        - :py:meth:`_EvalContext.attribute()`: Evaluate attribute values
          (has an additional argument `default` to provide values for missing attributes)
        - :py:meth:`_EvalContext.text()`: Evaluate text of tags
        - :py:meth:`_EvalContext.all_attributes()`: Evaluate all attributes of a tag
        - :py:meth:`_EvalContext.parent_attributes()`: Evaluate attribute values of parent of given tag
        - :py:meth:`_EvalContext.single_value()`: Evaluate `value` and `unit` attribute of tag
        - :py:meth:`_EvalContext.tag_exists()`: Evaluate whether a given tag exists
        - :py:meth:`_EvalContext.number_nodes()`: Evaluate how many elements of the tag are present
        - :py:meth:`_EvalContext.attribute_exists()`: Evaluate whether an attribute exists
        - :py:meth:`_EvalContext.simple_xpath()`: Evaluate the simple xpath expression for a given tag
        - *Nested Context* :py:meth:`_EvalContext.find()`: Find the first occurrence of the tag and provide a nested
          context to that element
        - *Nested Context* :py:meth:`_EvalContext.iter()`: Find all occurrences of the tag and provide a nested
          context to these elements when iterated over
        - *Nested Context* :py:meth:`_EvalContext.nested()`: Explicitly inherit all the context except the
          XML tree to a new context with the XML element replaced by the given argument

    Example Usage::

        from masci_tools.io.fleur_xml import load_inpxml, FleurXMLContext
        xmltree, schema_dict = load_inpxml('/path/to/inp.xml')

        with FleurXMLContext(xmltree, schema_dict) as root:
            spins = root.attribute('jspins')
            noco = root.attribute('l_noco', default=False)

            #Not nesting the context we need to specify which elements are meant
            mt_radii = root.attribute('radius', contains='species')

            #Nesting using find
            with root.find('atomspecies') as all_species:
                mt_radii = all_species.attribute('radius')

            #Nesting using iter
            mt_radii = []
            for species in root.iter('species'):
                mt_radii.append(species.attribute('radius'))

    """
    yield _EvalContext(etree_or_element, schema_dict, constants=constants, logger=logger)


def get_constants(xmltree: XMLLike | etree.XPathElementEvaluator,
                  schema_dict: fleur_schema.InputSchemaDict | fleur_schema.OutputSchemaDict,
                  logger: Logger | None = None) -> dict[str, float]:
    """
    Reads in the constants defined in the inp.xml
    and returns them combined with the predefined constants from
    fleur as a dictionary

    :param root: root of the etree of the inp.xml file
    :param schema_dict: schema_dictionary of the version of the file to read (inp.xml or out.xml)
    :param logger: logger object for logging warnings, errors

    :return: a python dictionary with all defined constants
    """
    from masci_tools.util.constants import FLEUR_DEFINED_CONSTANTS
    from masci_tools.io.parsers.fleur_schema import NoPathFound
    import copy

    defined_constants = copy.deepcopy(FLEUR_DEFINED_CONSTANTS)
    with FleurXMLContext(xmltree, schema_dict, logger=logger, constants=defined_constants) as root:

        try:
            root.tag_exists('constant')
        except NoPathFound:
            warnings.warn('Cannot extract custom constants for the given root. Assuming defaults')
            return defined_constants

        if not root.tag_exists('constant'):  #Avoid warnings for empty constants
            return defined_constants

        constants = root.all_attributes('constant')
        if constants['name'] is not None:
            if not isinstance(constants['name'], list):
                constants = {key: [val] for key, val in constants.items()}
            for name, value in zip(constants['name'], constants['value']):
                if name not in defined_constants:
                    defined_constants[name] = value
                else:
                    if logger is not None:
                        logger.error('Ambiguous definition of constant %s', name)
                    raise KeyError(f'Ambiguous definition of constant {name}')

    return defined_constants



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the masci-tools package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/masci-tools    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit                                        #
###############################################################################
"""
masci-tools is a collection of tools for materials science.
Mainly for the use with the Fleur (https://flapw.de) and juKKR (https://jukkr.fz-juelich.de) code
developed at the Forschungszentrum Jülich (see <http://judft.de>).

Contains IO and parsers utitlies for these codes and
vis contains wrappers of matplotlib functionality to visualize common material science data.
Plus wrappers of visualisation for aiida-fleur workflow nodes
"""
import logging

__copyright__ = ('Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany. '
                 'All rights reserved.')
__license__ = 'MIT license, see LICENSE.txt file.'
__version__ = '0.10.1'
__authors__ = 'The JuDFT team. Also see AUTHORS.txt file.'

logging.getLogger(__name__).addHandler(logging.NullHandler())



# pylint: disable=unused-import
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""This subpackage contains the tabulator class for the tabulator subpackage, which turns
properties of a collections of objects into a table.
"""

import abc as _abc
import typing as _typing

import pandas as _pd

from .recipes import Recipe


class Tabulator(_abc.ABC):
    """For tabulation of a collection of objects' (common) properties into a dict or dataframe.

    List of external implementations:

    - aiida-jutools/io `NodeTabulator` for nodes -> pandas DataFrame.

    TODO: increase memory performance:

    The following points are meant for the `NodeTabulator` implementation, but may be of interest for other
    implementations.

    - use optional dtypes from recipe (see TODO in Recipe) when building table. otherwise, e.g. when returning
      pandas dataframe, all columns will have dtype 'object' or 'float64' and the table won't fit into memory
      anymore very quickly.
    - internal storage format dict of lists while building must remain, but when finished, convert to dict
      of numpy arrays -> more memory efficient. for repeated tabulate() calls (to build larger table), have
      to adjust code to concatenate new lists to existing numpy arrays when finished.
    - change tabulate() signature: should not return table, only build it. another method (e.g. table @property
      getter) should return table and before del (delete) its inner storage (self._table) of it, because return
      will likely create a copy. that way can ~half the memory footprint.
    - when returning pandas dataframe, and recipe supplied no dtypes, use automatic downcasting to smallest dtype
      for numeric types (pd.to_numeric), and maybe 'categorical' for string coluns (if num unique values << num
      values). See pandas > Scaling to large datasets for more.
    - maybe add save option (or method) and read() method to read in tabulated table. for pandas, that allow a user
      to easily reuse the dtypes information from the recipe.
    """

    def __init__(self, recipe: Recipe = None, **kwargs):
        """Initialize a tabulator object.

        The attribute :py:attr:`~.recipe` defines *what* to extract from a set of objects and put them in a table (
        properties = column names of the table, one table row per object), with property transformations if needed.
        The tabulator on the other hand, knows *how* to extract those properties. It depends on the nature of the
        objects. Therefore, different implementations of the tabulator exist for different object types, and the
        base class is abstract.

        If no recipe is given, a default empty recipe will be used, and if :py:meth:`~tabulate` is called, the
        empty include list of the recipe will be replaced by an auto-generated extensive include list of object
        properties via the :py:meth:`~autolist` method.

        :param recipe: Optional recipe.
        :param kwargs: Additional keyword arguments for subclasses.
        """
        if not recipe:
            recipe = Recipe()
        self.recipe = recipe
        self._table_types = []
        self._table = None

    @_abc.abstractmethod
    def autolist(self, obj: _typing.Any, overwrite: bool = False, pretty_print: bool = False, **kwargs):
        """Auto-generate an include list of properties to be tabulated from a given object.

        This can serve as an overview for customized include and exclude lists.
        :param obj: An example object of a type compatible with the tabulator.
        :param overwrite: True: replace recipe list with the auto-generated list. False: Only if recipe list empty.
        :param pretty_print: True: Print the generated list in pretty format.
        :param kwargs: Additional keyword arguments for subclasses.
        """

    def clear(self):
        """Clear table if already tabulated."""
        self._table = None

    @property
    def table(self) -> _typing.Any:
        """The result table. None if :py:meth:`~tabulate` not yet called."""
        return self._table

    @_abc.abstractmethod
    def tabulate(self,
                 collection: _typing.Any,
                 table_type: _typing.Type = _pd.DataFrame,
                 append: bool = True,
                 column_policy: str = 'flat',
                 **kwargs) -> _typing.Optional[_typing.Any]:
        """Tabulate the common properties of a collection of objects.

        :param collection: collection of objects with same set of properties.
        :param table_type: Type of the tabulated data. Usually a pandas DataFrame or a dict.
        :param append: True: append to table if not empty. False: Overwrite table.
        :param column_policy: 'flat': Flat table, column names are last keys per keypath, name conflicts produce
                              warnings. 'flat_full_path': Flat table, column names are full keypaths,
                              'multiindex': table with MultiIndex columns (if pandas: `MultiIndex` columns), reflecting
                              the full properties' keypath hierarchies.
        :param kwargs: Additional keyword arguments for subclasses.
        :return: Tabulated objects' properties.
        """



"""
Commands for plotting
"""
from .root import cli
import click

from masci_tools.cmdline.utils import echo


@cli.group('plot')
def plot():
    """Commands for visualizing data"""


@plot.command('fleur-bands')
@click.argument('banddos-file', type=click.Path(exists=True))
@click.option('--weight', '-w', type=str, default=None)
@click.option('--backend', type=click.Choice(['matplotlib', 'mpl', 'bokeh']), default='matplotlib')
@click.option('--save', is_flag=True)
@click.option('--show', is_flag=True)
@click.option('--recipe',
              '-r',
              type=click.Choice(
                  ['FleurBands', 'FleurOrbcompBands', 'FleurjDOSBands', 'FleurSimpleBands', 'FleurMCDBands']),
              default='FleurBands')
def plot_fleur_banddos_bands(banddos_file, weight, recipe, backend, save, show):
    """
    Plot bandstructures from the banddos.hdf file from Fleur
    """
    from masci_tools.io.parsers.hdf5 import HDF5Reader
    from masci_tools.io.parsers.hdf5 import recipes

    from masci_tools.vis.fleur import plot_fleur_bands

    try:
        recipe = getattr(recipes, recipe)
    except AttributeError:
        echo.echo_critical(f'The recipe {recipe} does not exist')

    if not show and not save:
        show = True

    with HDF5Reader(banddos_file) as h5reader:
        data, attributes = h5reader.read(recipe=recipe)

    plot_fleur_bands(data, attributes, weight=weight, backend=backend, show=show, save_plots=save)


@plot.command('fleur-dos')
@click.argument('banddos-file', type=click.Path(exists=True))
@click.option('--total', type=bool, default=True)
@click.option('--interstitial', type=bool, default=True)
@click.option('--atoms', default='all')
@click.option('--l_resolved', default=None)
@click.option('--backend', type=click.Choice(['matplotlib', 'mpl', 'bokeh']), default='matplotlib')
@click.option('--save', is_flag=True)
@click.option('--show', is_flag=True)
@click.option('--recipe',
              '-r',
              type=click.Choice(['FleurDOS', 'FleurORBCOMP', 'FleurJDOS', 'FleurMCD']),
              default='FleurDOS')
def plot_fleur_banddos_dos(banddos_file, total, interstitial, atoms, l_resolved, recipe, backend, save, show):
    """
    Plot density of states from the banddos.hdf file from Fleur
    """
    from masci_tools.io.parsers.hdf5 import HDF5Reader
    from masci_tools.io.parsers.hdf5 import recipes

    from masci_tools.vis.fleur import plot_fleur_dos

    try:
        recipe = getattr(recipes, recipe)
    except AttributeError:
        echo.echo_critical(f'The recipe {recipe} does not exist')
    if not show and not save:
        show = True

    with HDF5Reader(banddos_file) as h5reader:
        data, attributes = h5reader.read(recipe=recipe)

    plot_fleur_dos(data,
                   attributes,
                   show_total=total,
                   show_interstitial=interstitial,
                   show_atoms=atoms,
                   show_lresolved=l_resolved,
                   backend=backend,
                   show=show,
                   save_plots=save)



"""
This module defines some aliases used in typing
"""
from __future__ import annotations

from typing import TypeVar, Any, IO, Union
import sys
if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias
from lxml import etree
from pathlib import Path
import os

#Type for xml setters/getters for xpath objects that can be used in `eval_xpath`
from masci_tools.util.xml.xpathbuilder import XPathBuilder

XPathLike: TypeAlias = Union[str, bytes, etree.XPath, XPathBuilder]
"""
Type for xpath expressions
"""

TXPathLike = TypeVar('TXPathLike', bound=XPathLike)
"""
Type for xpath expressions
"""

FileLike: TypeAlias = Union[str, bytes, Path, os.PathLike, IO[Any]]
"""
Type used for functions accepting file-like objects, i.e. handles or file paths
"""

XMLFileLike: TypeAlias = Union[etree._ElementTree, etree._Element, FileLike]
"""
Type used for functions accepting xml-file-like objects, i.e. handles or file paths
or already parsed xml objects
"""

XMLLike: TypeAlias = Union[etree._Element, etree._ElementTree]
"""
Type used for functions accepting xml objects from lxml
"""



"""
This module implements a commandline tool available with ``masci-tools inpxml`` to
convert inp.xml files between different file versions
"""
from __future__ import annotations

from collections import defaultdict
from typing import Iterable, NamedTuple, Sequence, TypeVar, cast
try:
    from typing import TypedDict
except ImportError:
    from typing_extensions import TypedDict
import json
from pathlib import Path
import os

from lxml import etree
import click
import tabulate

from masci_tools.io.parsers.fleur_schema import InputSchemaDict
from masci_tools.io.fleur_xml import load_inpxml
from masci_tools.util.schema_dict_util import evaluate_attribute, tag_exists
from masci_tools.util.typing import XMLLike, FileLike
from masci_tools.util.xml.xml_setters_basic import xml_delete_tag, xml_delete_att, xml_create_tag, _reorder_tags, xml_set_attrib_value_no_create
from masci_tools.util.xml.xml_setters_names import set_attrib_value
from masci_tools.util.xml.common_functions import split_off_attrib, split_off_tag, eval_xpath
from masci_tools.cmdline.parameters.slice import ListElement
from masci_tools.cmdline.utils import echo

FILE_DIRECTORY = Path(__file__).parent.resolve()


#These are the possible conversions that we can do
class MoveAction(NamedTuple):
    """
    NamedTuple representing the action of moving a tag/attribute to a different place.
    The name of this tag/attribute can also be changed
    """
    old_name: str
    new_name: str
    old_path: str
    new_path: str
    attrib: bool = False

    @classmethod
    def from_path(cls, old: str, new: str) -> MoveAction:
        """
        Construct a MoveAction from two given xpaths

        :param old: old str of the xpath
        :param new: new str of the xpath
        """

        attrib = '@' in old
        if attrib:
            old_path, old_name = split_off_attrib(old)
            new_path, new_name = split_off_attrib(new)
        else:
            old_path, new_path = old, new
            _, old_name = split_off_tag(old_path)
            _, new_name = split_off_tag(new_path)

        return cls(old_name=old_name, new_name=new_name, old_path=old_path, new_path=new_path, attrib=attrib)

    @classmethod
    def from_remove_and_create(cls, remove: RemoveAction, create: CreateAction) -> MoveAction:
        """
        Construct a MoveAction from a remove and create action, merging them together

        :param remove: RemoveAction to merge
        :param create: CreateAction to merge
        """

        if remove.attrib != create.attrib:
            raise ValueError('Inconsistent Remove/Create Actions')

        return cls(old_name=remove.name,
                   new_name=create.name,
                   old_path=remove.path,
                   new_path=create.path,
                   attrib=remove.attrib)


class NormalizedMoveAction(NamedTuple):
    """
    NamedTuple representing the action of moving a tag/attribute to a different place.
    The name of this tag/attribute can also be changed

    This MoveAction holds an additional attribute of an intermediate xpath to use, since the
    tag could not be in the expected starting position if other tags are moved beforehand
    """
    old_name: str
    new_name: str
    old_path: str
    new_path: str
    actual_path: str
    attrib: bool = False

    @classmethod
    def from_move(cls, move_action: MoveAction, actual_path: str) -> NormalizedMoveAction:
        """
        Construct a NormalizedMoveAction from a given move and the additional actual xpath to use

        :param move_action: MoveAction to construct the object from
        :param actual_path: str of the intermediate xpath to use as a starting point
        """

        return cls(**move_action._asdict(), actual_path=actual_path)


class AmbiguousAction(NamedTuple):
    """
    NamedTuple representing a change in paths that cannot be resolved automatically
    """
    name: str
    old_paths: tuple[str, ...]
    new_paths: tuple[str, ...]
    attrib: bool

    @classmethod
    def from_paths(cls, old: Iterable[str], new: Iterable[str]) -> AmbiguousAction:
        """
        Construct AmbiguousAction from given paths

        :param old: Iterable of the paths in the old version
        :param new: Iterable of paths in the new version
        """
        path = list(old)[0]
        attrib = '@' in path
        if attrib:
            _, name = split_off_attrib(path)
        else:
            _, name = split_off_tag(path)

        return cls(name=name, old_paths=tuple(old), new_paths=tuple(new), attrib=attrib)


class RemoveAction(NamedTuple):
    """
    NamedTuple representing the action of removing a tag/attribute from the xml tree
    """
    name: str
    path: str
    warning: str = ''
    attrib: bool = False

    @classmethod
    def from_path(cls, xpath: str) -> RemoveAction:
        """
        Construct a RemoveAction from a given xpath

        :param xpath: str of the xpath
        """

        attrib = '@' in xpath
        if attrib:
            path, name = split_off_attrib(xpath)
        else:
            path = xpath
            _, name = split_off_tag(xpath)

        return cls(name=name, path=path, attrib=attrib)


class CreateAction(NamedTuple):
    """
    NamedTuple representing the action of creating a tag/attribute in the xml tree
    """
    name: str
    path: str
    attrib: bool = False
    element: str | None = None

    @classmethod
    def from_path(cls, xpath: str, element: str | None = None) -> CreateAction:
        """
        Construct a CreateAction from a given xpath

        :param xpath: str of the xpath
        """

        attrib = '@' in xpath
        if attrib:
            path, name = split_off_attrib(xpath)
        else:
            path = xpath
            _, name = split_off_tag(xpath)

        return cls(name=name, path=path, attrib=attrib, element=element)


SimpleAction = TypeVar('SimpleAction', CreateAction, RemoveAction)
NamedAction = TypeVar('NamedAction', CreateAction, RemoveAction, AmbiguousAction)


class Actions(TypedDict):
    remove: list[RemoveAction]
    create: list[CreateAction]
    move: list[MoveAction | NormalizedMoveAction]


class FileConversion(TypedDict, total=False):
    from_version: str
    to_version: str
    tag: Actions
    attrib: Actions


def analyse_paths(
    schema_start: InputSchemaDict, schema_target: InputSchemaDict, path_entries: str | list[str]
) -> tuple[list[RemoveAction], list[CreateAction], list[MoveAction], list[AmbiguousAction]]:
    """
    Gather the initial path differences between the two given input schema dictionaries
    for the given entries. If multiple enetries are given they are first merged together

    :param schema_start: InputSchemaDict to start from
    :param schema_start: InputSchemaDict to end up at
    :param path_entries: which entries of the schema dictionary should be analysed

    :returns: Tuple of List of Actions put into four categories (remove, create, move, ambiguous)
    """

    if not isinstance(path_entries, list):
        path_entries = [path_entries]

    paths_start = defaultdict(list)
    paths_target = defaultdict(list)
    for path_entry in path_entries:
        new_paths = schema_start[path_entry]
        for name, paths in new_paths.items():
            if isinstance(paths, str):
                paths = [paths]
            paths_start[name].extend(paths)
        new_paths = schema_target[path_entry]
        for name, paths in new_paths.items():
            if isinstance(paths, str):
                paths = [paths]
            paths_target[name].extend(paths)

    removed_keys = paths_start.keys() - paths_target.keys()
    remove: list[RemoveAction] = []
    for key in removed_keys:
        remove.extend(RemoveAction.from_path(path) for path in paths_start[key])

    new_keys = paths_target.keys() - paths_start.keys()
    create: list[CreateAction] = []
    for key in new_keys:
        create.extend(CreateAction.from_path(path) for path in paths_target[key])

    move = []
    ambiguous = []
    possible_change_keys = paths_start.keys() & paths_target.keys()
    for key in possible_change_keys:
        if paths_start[key] == paths_target[key]:
            continue

        old_paths = set(paths_start[key])
        new_paths = set(paths_target[key])

        different_paths = old_paths.symmetric_difference(new_paths)
        if len(different_paths) == 1:
            path = different_paths.pop()
            if path in old_paths:
                remove.append(RemoveAction.from_path(path))
            else:
                create.append(CreateAction.from_path(path))
        elif len(different_paths) == 2:
            if all(path in old_paths for path in different_paths):
                remove.extend(RemoveAction.from_path(path) for path in different_paths)
                continue
            if all(path in new_paths for path in different_paths):
                create.extend(CreateAction.from_path(path) for path in different_paths)
                continue
            first_path = different_paths.pop()
            second_path = different_paths.pop()
            if first_path in old_paths:
                move.append(MoveAction.from_path(old=first_path, new=second_path))
            else:
                move.append(MoveAction.from_path(old=second_path, new=first_path))
        else:
            ambiguous.append(AmbiguousAction.from_paths(old=old_paths, new=new_paths))

    return remove, create, move, ambiguous


def resolve_ambiguouities(ambiguous: list[AmbiguousAction],
                          remove: list[RemoveAction],
                          create: list[CreateAction],
                          move: list[MoveAction],
                          remove_move: bool = False,
                          tag_remove: list[RemoveAction] | None = None,
                          tag_move: list[MoveAction] | None = None) -> None:
    """
    Try to resolve ambuouities by using additional information from moved/created removed tags/attributes
    """

    for action in ambiguous.copy():

        ambiguous.remove(action)

        old_paths = set(action.old_paths)
        new_paths = set(action.new_paths)

        if tag_remove is None:
            tag_remove = remove
        for old_path in old_paths.copy():
            for action2 in tag_remove:
                if action2.path in old_path:
                    old_paths.discard(old_path)

        if remove_move:
            if tag_move is None:
                tag_move = move
            for old_path in old_paths.copy():
                for new_path in new_paths.copy():
                    for action_move in tag_move:
                        if old_path == new_path:
                            continue
                        if old_path.replace(action_move.old_path, '') == new_path.replace(action_move.new_path, ''):
                            old_paths.discard(old_path)
                            new_paths.discard(new_path)

        if old_paths == new_paths:
            continue

        different_paths = old_paths.symmetric_difference(new_paths)

        if all(path in new_paths for path in different_paths):
            create.extend(CreateAction.from_path(path) for path in different_paths)
            continue

        if all(path in old_paths for path in different_paths):
            remove.extend(RemoveAction.from_path(path) for path in different_paths)
            continue

        if len(different_paths) == 1:
            path = different_paths.pop()
            if path in old_paths:
                remove.append(RemoveAction.from_path(path))
            else:
                create.append(CreateAction.from_path(path))
        elif len(different_paths) == 2:
            if all(path in old_paths for path in different_paths):
                remove.extend(RemoveAction.from_path(path) for path in different_paths)
                continue
            if all(path in new_paths for path in different_paths):
                create.extend(CreateAction.from_path(path) for path in different_paths)
                continue
            first_path = different_paths.pop()
            second_path = different_paths.pop()
            if first_path in old_paths:
                move.append(MoveAction.from_path(old=first_path, new=second_path))
            else:
                move.append(MoveAction.from_path(old=second_path, new=first_path))
        else:
            ambiguous.append(AmbiguousAction.from_paths(old=old_paths, new=new_paths))


def trim_paths(paths: list[SimpleAction]) -> list[SimpleAction]:

    path_copy = paths.copy()
    for action in path_copy:
        for action2 in path_copy:
            if action == action2:
                continue
            if action.path in action2.path:
                if action2 in paths:
                    paths.remove(action2)

    return paths


def trim_attrib_paths(paths: list[SimpleAction], tag_paths: list[SimpleAction]) -> list[SimpleAction]:

    path_copy = paths.copy()
    for tag_action in tag_paths:
        for attrib_action in path_copy:
            if tag_action.path in attrib_action.path:
                if attrib_action in paths:
                    paths.remove(attrib_action)

    return paths


def trim_move_paths(paths: list[MoveAction]) -> list[MoveAction]:

    path_copy = paths.copy()
    for action in path_copy:
        for action2 in path_copy:
            if action == action2:
                continue
            if action.old_path in action2.old_path:
                if action2.old_path.replace(action.old_path, '') == action2.new_path.replace(action.new_path, ''):
                    if action2 in paths:
                        paths.remove(action2)

    return paths


def trim_attrib_move_paths(paths: list[MoveAction], tag_paths: list[MoveAction]) -> list[MoveAction]:

    path_copy = paths.copy()
    for action in tag_paths:
        for action2 in path_copy:
            if action.old_path in action2.old_path:
                if action2.old_path.replace(action.old_path, '') == action2.new_path.replace(action.new_path, ''):
                    if action2 in paths:
                        paths.remove(action2)

    return paths


def remove_action(actions: list[NamedAction], name: str) -> list[NamedAction]:
    """
    Find all the occurrences of actions with a given name and return them and remove them from the list

    :param actions: List of actions
    :param name: str name to find

    :returns: List of actions matching the given name
    """

    matching = [action for action in actions if action.name == name]
    for action in matching:
        actions.remove(action)

    if not matching:
        raise ValueError(f'Action {name} not found')

    return matching


def echo_actions(actions: Sequence[NamedTuple], ignore: Iterable[str] | None = None, header: str = '') -> None:
    """
    Echo a list of actions in a nicely formatted list

    :param actions: list of actions to show
    :param ignore: Iterable of str with attributes to ignore in the table
    :param header: str optional Title for the table
    """

    if ignore is None:
        ignore = {'attrib'}

    actions_dict = [action._asdict() for action in actions]

    if ignore:
        for action in actions_dict:
            for key in ignore:
                action.pop(key, None)

    click.echo(header)
    if any(isinstance(action, AmbiguousAction) for action in actions):
        first = True
        for action in actions_dict:
            old_length = len(action['old_paths'])
            new_length = len(action['new_paths'])
            if old_length < new_length:
                action['old_paths'] += (None,) * (new_length - old_length)
            elif new_length < old_length:
                action['new_paths'] += (None,) * (old_length - new_length)
            if 'name' in action:
                action['name'] = [action['name']] + [None] * max(new_length, old_length)
            if first:
                click.echo(tabulate.tabulate(action, headers=['Name', 'Old paths', 'New paths']))
                first = False
            else:
                click.echo(tabulate.tabulate(action))
    else:
        click.echo(tabulate.tabulate(actions_dict, showindex=True, headers='keys'))
    click.echo()


def load_conversion(from_version: str, to_version: str) -> FileConversion:
    """
    Load the conversion between the given versions from a stored json file

    :param from_version: str of the initial version
    :param to_version: str of the final version

    :returns: a dict with the actions to perform
    """
    filepath = FILE_DIRECTORY / 'conversions' / f"conversion_{from_version.replace('.','')}_to_{to_version.replace('.','')}.json"

    with open(filepath, encoding='utf-8') as f:
        conversion: FileConversion = json.load(f)

    #convert back to the namedtuples
    conversion['tag']['create'] = [CreateAction(*action) for action in conversion['tag']['create']]
    conversion['tag']['remove'] = [RemoveAction(*action) for action in conversion['tag']['remove']]
    move: list[MoveAction | NormalizedMoveAction] = []
    for action in conversion['tag']['move']:
        if len(action) == 6:
            move.append(NormalizedMoveAction(*action))
        else:
            move.append(MoveAction(*action))
    conversion['tag']['move'] = move

    conversion['attrib']['create'] = [CreateAction(*action) for action in conversion['attrib']['create']]
    conversion['attrib']['remove'] = [RemoveAction(*action) for action in conversion['attrib']['remove']]
    move = []
    for action in conversion['attrib']['move']:
        if len(action) == 6:
            move.append(NormalizedMoveAction(*action))
        else:
            move.append(MoveAction(*action))
    conversion['attrib']['move'] = move

    return conversion


def dump_conversion(conversion: FileConversion) -> None:
    """
    Save the given conversion as a json file in the conversions subfolder

    :param conversion: dict representing the conversion
    """

    filepath = FILE_DIRECTORY / 'conversions' / f"conversion_{conversion['from_version'].replace('.','')}_to_{conversion['to_version'].replace('.','')}.json"
    os.makedirs(filepath.parent, exist_ok=True)

    #Drop all create actions, which have no element set
    conversion['tag']['create'] = [action for action in conversion['tag']['create'] if action.element is not None]
    conversion['attrib']['create'] = [action for action in conversion['attrib']['create'] if action.element is not None]

    with open(filepath, 'w', encoding='utf-8') as f:
        json.dump(conversion, f, indent=2, sort_keys=False)


def _rename_elements(remove: list[RemoveAction], create: list[CreateAction], move: list[MoveAction], from_version: str,
                     to_version: str, name: str) -> tuple[list[RemoveAction], list[CreateAction], list[MoveAction]]:
    """
    Get user input on tags that were renamed
    """

    remove = sorted(remove, key=lambda x: x.name)
    create = sorted(create, key=lambda x: x.name)

    rename = True
    while remove and create and rename:
        echo_actions(remove,
                     header=f'The following {name} are not found in the target version:',
                     ignore={'attrib', 'warning'})
        echo_actions(create,
                     header=f'The following {name} are not found in the start version:',
                     ignore={'attrib', 'element'})

        rename = click.confirm(f'Are there {name} that were renamed?')

        if rename:
            old_name = click.prompt(f'Name in version {from_version}',
                                    type=click.Choice([action.name for action in remove], case_sensitive=False),
                                    show_choices=False)
            new_name = click.prompt(f'Name in version {to_version}',
                                    type=click.Choice([action.name for action in create], case_sensitive=False),
                                    show_choices=False)

            remove_list = remove_action(remove, old_name)
            create_list = remove_action(create, new_name)

            if len(remove_list) != len(create_list):
                raise ValueError('Not supported')
            move.extend(
                MoveAction.from_remove_and_create(remove, create) for remove, create in zip(remove_list, create_list))
            move = trim_move_paths(move)
    return remove, create, move


def _create_tag_elements(create: list[CreateAction], to_schema: InputSchemaDict) -> list[CreateAction]:
    """
    Get user input on tags to create
    """
    create_prompt = True
    while create and create_prompt:
        echo_actions(create, header='The following tags will be created:')

        create_prompt = click.confirm('Do you want to set an element to create?')

        if create_prompt:
            name = click.prompt('Name of the element',
                                type=click.Choice([action.name for action in create], case_sensitive=False),
                                show_choices=False)
            create_actions = remove_action(create, name)

            if len(create_actions) != 1:
                raise ValueError('Not implemented')
            create_action = create_actions[0]
            allowed_attribs = to_schema['tag_info'][create_action.path]['attribs']

            attribs = {}
            for attrib in allowed_attribs:
                value = click.prompt(f'Value for {attrib}', default=None)
                if value is not None:
                    attribs[attrib] = value

            elem = etree.tostring(etree.Element(create_action.name, **attribs), encoding='unicode', pretty_print=True)
            echo.echo_info(f'The following element will be created: {elem}')

            create.append(create_action._replace(element=elem))
        create = sorted(create, key=lambda x: x.name)
    return create


def _add_warnings_on_remove(remove: list[RemoveAction], name: str) -> list[RemoveAction]:
    """
    Get user input on warnings to show
    """
    warning_prompt = True
    while remove and warning_prompt:
        echo_actions(remove, header=f'The following {name} will be removed:')

        warning_prompt = click.confirm(f'Do you want to add a warning if a {name} is removed?')

        if warning_prompt:
            elem_name = click.prompt('Name of the element',
                                     type=click.Choice([action.name for action in remove], case_sensitive=False),
                                     show_choices=False)
            remove_list = remove_action(remove, elem_name)
            while click.confirm('Do you want to enter another element?', default=False):
                elem_name = click.prompt('Name of the element',
                                         type=click.Choice([action.name for action in remove], case_sensitive=False),
                                         show_choices=False)
                remove_list += remove_action(remove, elem_name)

            warning = click.prompt('Enter the warning message', type=click.types.StringParamType())
            remove.extend(action._replace(warning=warning) for action in remove_list)
        remove = sorted(remove, key=lambda x: x.name)
    return remove


def _create_attrib_elements(create: list[CreateAction], to_schema: InputSchemaDict) -> list[CreateAction]:
    """
    Get user input on attributes to create
    """
    create_prompt = True
    while create and create_prompt:
        echo_actions(create, header='The following attributes will be created:')

        create_prompt = click.confirm('Do you want to set a value for a given attribute?')

        if create_prompt:
            name = click.prompt('Name of the attribute',
                                type=click.Choice([action.name for action in create], case_sensitive=False),
                                show_choices=False)
            create_actions = remove_action(create, name)

            if len(create_actions) != 1:
                raise ValueError('Not implemented')
            create_action = create_actions[0]
            default = to_schema['tag_info'][create_action.path]['optional_attribs'].get(create_action.name)
            if default is not None:
                default = str(default)

            value = click.prompt(f'Value for {create_action.name}', default=default)
            if value is not None:
                create.append(create_action._replace(element=value))
        create = sorted(create, key=lambda x: x.name)
    return create


def _manual_resolution(ambiguous: list[AmbiguousAction], remove: list[RemoveAction], create: list[CreateAction],
                       move: list[MoveAction],
                       name: str) -> tuple[list[RemoveAction], list[CreateAction], list[MoveAction]]:
    """
    Prompt the user for input on actions that cannot be determined automagically
    """

    ambiguous = sorted(ambiguous, key=lambda x: x.name)

    while ambiguous:
        echo_actions(ambiguous, header=f'The following {name} could not be resolved automatically:')

        if len(ambiguous) > 1:
            name = click.prompt('Enter the name you want to clarify',
                                type=click.Choice([action.name for action in ambiguous], case_sensitive=False),
                                show_choices=False)

            entries = remove_action(ambiguous, name)

            if len(entries) != 1:
                raise NotImplementedError("It's broken :-(")
            entry = entries[0]
        else:
            entry = ambiguous.pop(0)

        click.echo(f'Entry {entry.name}:')

        old_paths = sorted(entry.old_paths)
        new_paths = sorted(entry.new_paths)

        while (new_paths or old_paths) and new_paths != old_paths:
            old_paths_display, new_paths_display = old_paths.copy(), new_paths.copy()
            if len(old_paths) < len(new_paths):
                old_paths_display += [None] * (len(new_paths) - len(old_paths))  #type:ignore
            elif len(new_paths) < len(old_paths):
                new_paths_display += [None] * (len(old_paths) - len(new_paths))  #type:ignore

            click.echo(
                tabulate.tabulate(list(zip(old_paths_display, new_paths_display)),
                                  showindex=True,
                                  headers=['Old Paths', 'New Paths']))

            action = click.prompt('Which action should be performed',
                                  type=click.Choice(['create', 'remove', 'move'], case_sensitive=False))
            if action == 'remove':
                paths = click.prompt('Enter the row you want to remove from the old paths',
                                     type=ListElement(old_paths, return_list=True))
                for path in paths:
                    remove.append(RemoveAction.from_path(path))
                    old_paths.remove(path)
            elif action == 'create':
                paths = click.prompt('Enter the row you want to create from the new paths',
                                     type=ListElement(new_paths, return_list=True))
                for path in paths:
                    create.append(CreateAction.from_path(path))
                    new_paths.remove(path)
            elif action == 'move':
                old_path_row = click.prompt('Enter the row you want to remove from the old paths',
                                            type=click.types.IntParamType())
                new_path_row = click.prompt('Enter the row you want to remove from the new paths',
                                            type=click.types.IntParamType())
                old_path = old_paths.pop(old_path_row)
                new_path = new_paths.pop(new_path_row)
                move.append(MoveAction.from_path(old=old_path, new=new_path))

        echo.echo_success('Ambiguouity successfully resolved')
    return remove, create, move


def _xml_create_tag_with_parents(xmltree: XMLLike, xpath: str, node: etree._Element) -> None:
    """
    Create a tag at the given xpath together with it's parents if they are missing
    xml_create_tag cannot create subtags, but since we know that we have simple xpaths
    we can do it here

    No tag order is enforced here, since we are in intermediate steps

    :param xmltree: etree ElementTree to operate on
    :param xpath: xpath of the parent node
    :param node: node to create
    """

    parent_nodes = eval_xpath(xmltree, xpath, list_return=True)
    to_create = []
    while not parent_nodes:
        parent_path, parent_name = split_off_tag(xpath)
        to_create.append((parent_path, parent_name))
        parent_nodes = eval_xpath(xmltree, parent_path, list_return=True)

    for parent_path, name in reversed(to_create):
        xml_create_tag(xmltree, parent_path, name)

    xml_create_tag(xmltree, xpath, node)


def _xml_delete_attribute_with_warnings(xmltree: XMLLike, xpath: str, name: str, warning: str = '') -> None:
    """
    Delete the attribute at the given xpath with the given name and show a warning if one is given

    :param xmltree: etree ElementTree to operate on
    :param xpath: xpath of the parent node
    :param name: name of the attribute
    :param warning: str of the warning to show in case there are nodes that are removed
    """
    values = eval_xpath(xmltree, f'{xpath}/@{name}', list_return=True)
    if values:
        if warning:
            echo.echo_warning(warning)
        xml_delete_att(xmltree, xpath, name)


def _xml_delete_tag_with_warnings(xmltree: XMLLike, xpath: str, warning: str = '') -> None:
    """
    Delete the tag at the given xpath and show a warning if one is given

    :param xmltree: etree ElementTree to operate on
    :param xpath: xpath of the node
    :param warning: str of the warning to show in case there are nodes that are removed
    """
    nodes = eval_xpath(xmltree, xpath, list_return=True)
    if nodes:
        if warning:
            echo.echo_warning(warning)
        xml_delete_tag(xmltree, xpath)


def _reorder_tree(parent: etree._Element, schema_dict: InputSchemaDict, base_xpath: str | None = None) -> None:
    """
    Order the elements to be in the correct order for the given schema_dict
    """

    #Check if this is the first call to this routine
    if base_xpath is None:
        base_xpath = f'/{parent.tag}'

    for element in parent:
        new_base_xpath = f'{base_xpath}/{element.tag}'
        _reorder_tree(element, schema_dict, base_xpath=new_base_xpath)

    if base_xpath in schema_dict['tag_info']:
        order = schema_dict['tag_info'][base_xpath]['order']
    else:
        order = []

    if order:
        parent = _reorder_tags(parent, order)


def convert_inpxml(xmltree: etree._ElementTree, schema_dict: InputSchemaDict, to_version: str) -> etree._ElementTree:
    """
    Convert the given xmltree to the given file version

    :param xmltree: XML tree to convert
    :param schema_dict: SchemaDict corresponding to the original file version
    :param to_version: file version to which to convert

    :returns: the XML tree converted to the given file version
    """
    schema_dict_target = InputSchemaDict.fromVersion(to_version)

    #We want to leave comments in so we cannot use clear_xml for the xinclude feature
    #Here we just include and write out the complete xml file
    xmltree.xinclude()
    from_version = evaluate_attribute(xmltree, schema_dict, 'fleurInputVersion')

    conversion = load_conversion(from_version, to_version)

    set_attrib_value(xmltree, schema_dict_target, 'fleurInputVersion', to_version)

    action: MoveAction | RemoveAction | CreateAction | NormalizedMoveAction
    #Collect the nodes to be moved
    move_tags: list[list[etree._Element]] = []
    for action in conversion['tag']['move']:
        move_tags.append(eval_xpath(xmltree, action.old_path, list_return=True))  #type: ignore

    #Collect the attributes to be moved
    move_attribs = []
    for action in conversion['attrib']['move']:
        move_attribs.append(eval_xpath(xmltree, f'{action.old_path}/@{action.old_name}', list_return=True))

    for action in conversion['tag']['remove']:
        _xml_delete_tag_with_warnings(xmltree, action.path, warning=action.warning)

    for action in conversion['attrib']['remove']:
        _xml_delete_attribute_with_warnings(xmltree, action.path, action.name, warning=action.warning)

    for action, nodes in zip(conversion['tag']['move'], move_tags):
        if isinstance(action, NormalizedMoveAction):
            path = action.actual_path
        else:
            path = action.old_path

        if eval_xpath(xmltree, path, list_return=True):
            xml_delete_tag(xmltree, path)

        for node in nodes:
            path, _ = split_off_tag(action.new_path)
            node.tag = action.new_name
            #Order is not kept here (it is corrected at the end)
            _xml_create_tag_with_parents(xmltree, path, node)

    for action in conversion['tag']['create']:
        if action.element is not None:
            path, _ = split_off_tag(action.path)
            node = etree.fromstring(action.element)
            _xml_create_tag_with_parents(xmltree, path, node)

    for action, values in zip(conversion['attrib']['move'], move_attribs):
        if isinstance(action, NormalizedMoveAction):
            path = action.actual_path
        else:
            path = action.old_path

        if eval_xpath(xmltree, f'{path}/@{action.old_name}', list_return=True):
            xml_delete_att(xmltree, path, action.old_name)
        if values:
            nodes: list[etree._Element] = eval_xpath(xmltree, action.new_path, list_return=True)  #type: ignore
            if nodes:
                xml_set_attrib_value_no_create(xmltree, action.new_path, action.new_name, values)

    for action in conversion['attrib']['create']:
        if action.element is not None:
            nodes: list[etree._Element] = eval_xpath(xmltree, action.path, list_return=True)  #type: ignore
            if nodes:
                xml_set_attrib_value_no_create(xmltree, action.path, action.name, action.element)

    _reorder_tree(xmltree.getroot(), schema_dict_target)
    schema_dict_target.validate(xmltree)

    #If there was no relax.xml included we need to rewrite the xinclude tag for it
    INCLUDE_NSMAP = {'xi': 'http://www.w3.org/2001/XInclude'}
    INCLUDE_TAG = etree.QName(INCLUDE_NSMAP['xi'], 'include')
    FALLBACK_TAG = etree.QName(INCLUDE_NSMAP['xi'], 'fallback')
    if not tag_exists(xmltree, schema_dict_target, 'relaxation'):
        xinclude_elem = etree.Element(INCLUDE_TAG, href='relax.xml', nsmap=INCLUDE_NSMAP)
        xinclude_elem.append(etree.Element(FALLBACK_TAG))
        xmltree.getroot().append(xinclude_elem)

    etree.indent(xmltree)
    return xmltree


@click.group('inpxml')
def inpxml():
    """
    Tool for converting inp.xml files to different versions
    """


@inpxml.command('convert')
@click.argument('xml-file', type=click.Path(exists=True))
@click.argument('to_version', type=str)
@click.option('--output-file', '-o', type=str, default='inp.xml', help='Name of the output file')
@click.option('--overwrite', is_flag=True, help='If the flag is given and the file already exists it is overwritten')
@click.pass_context
def cmd_convert_inpxml(ctx: click.Context, xml_file: FileLike, to_version: str, output_file: str,
                       overwrite: bool) -> None:
    """
    Convert the given XML_FILE file to version TO_VERSION

    XML_FILE is the file to convert
    TO_VERSION is the file version of the finale input file
    """
    xmltree, schema_dict = load_inpxml(xml_file)
    from_version = evaluate_attribute(xmltree, schema_dict, 'fleurInputVersion')

    try:
        load_conversion(from_version, to_version)
    except FileNotFoundError:
        echo.echo_warning(f'No conversion available between versions {from_version} to {to_version}')
        if click.confirm('Do you want to generate this conversion now'):
            ctx.invoke(generate_inp_conversion, from_version=from_version, to_version=to_version)
        else:
            echo.echo_critical('Cannot convert')

    if Path(output_file).is_file():
        if not overwrite:
            echo.echo_critical(f'The output file {output_file} already exists. Use the overwrite flag to ignore')
        echo.echo_warning(f'The output file {output_file} already exists. Will be overwritten')

    try:
        convert_inpxml(xmltree, schema_dict, to_version)
        echo.echo_success('The conversion was successful')
        echo.echo_info(
            'It is not guaranteed that a FLEUR calculation will behave in the exact same way as the old input file\n'
            'Please check the file for correctness beforehand')
    except etree.DocumentInvalid as err:
        echo.echo_critical(
            f'inp.xml conversion did not finish successfully. The resulting file violates the XML schema with:\n {err}')

    xmltree.write(output_file, encoding='utf-8', pretty_print=True)
    echo.echo_success(f'Converted file written to {output_file}')


@inpxml.command('generate-conversion')
@click.argument('from_version', type=str)
@click.argument('to_version', type=str)
@click.option('--show/--no-show', default=True, help='Show a summary of the conversion at the end')
@click.pass_context
def generate_inp_conversion(ctx: click.Context, from_version: str, to_version: str, show: bool) -> FileConversion:
    """
    Generate the conversions from FROM_VERSION to TO_VERSION

    FROM_VERSION is the file version of the initial input file
    TO_VERSION is the file version of the finale input file
    """
    TAG_ENTRY = 'tag_paths'
    ATTRIB_ENTRIES = ['unique_attribs', 'unique_path_attribs', 'other_attribs']

    from_schema = InputSchemaDict.fromVersion(from_version)
    to_schema = InputSchemaDict.fromVersion(to_version)

    remove_tags, create_tags, move_tags, ambiguous_tags = analyse_paths(from_schema, to_schema, TAG_ENTRY)

    remove_tags = trim_paths(remove_tags)
    create_tags = trim_paths(create_tags)
    move_tags = trim_move_paths(move_tags)

    remove_tags, create_tags, move_tags = _rename_elements(remove_tags, create_tags, move_tags, from_version,
                                                           to_version, 'tags')

    #Check again if we can now resolve ambiguouities
    resolve_ambiguouities(ambiguous_tags, remove_tags, create_tags, move_tags)
    remove_tags = trim_paths(remove_tags)
    create_tags = trim_paths(create_tags)
    move_tags = trim_move_paths(move_tags)

    remove_tags = sorted(remove_tags, key=lambda x: x.name)
    create_tags = sorted(create_tags, key=lambda x: x.name)
    move_tags = sorted(move_tags, key=lambda x: x.new_name)

    remove_tags, create_tags, move_tags = _manual_resolution(ambiguous_tags, remove_tags, create_tags, move_tags,
                                                             'tags')

    remove_tags = sorted(remove_tags, key=lambda x: x.name)
    create_tags = sorted(create_tags, key=lambda x: x.name)
    move_tags = sorted(move_tags, key=lambda x: x.new_name)

    #Make move_tags consistent
    for indx, action in enumerate(move_tags):
        #When the tag has been moved all paths afterward have to be adjusted
        if indx == len(move_tags) - 1:
            continue
        for indx_after, action_after in enumerate(move_tags[indx + 1:]):
            if action.old_path in action_after.old_path:
                intermediate_path = action_after.old_path.replace(action.old_path, action.new_path)
                #yapf:disable
                move_tags[indx_after + indx + 1] = NormalizedMoveAction.from_move(action_after, actual_path=intermediate_path)  #type:ignore
                #yapf:enable

    create_tags = _create_tag_elements(create_tags, to_schema)
    remove_tags = _add_warnings_on_remove(remove_tags, 'tags')

    remove_attrib, create_attrib, move_attrib, ambiguous_attrib = analyse_paths(from_schema, to_schema, ATTRIB_ENTRIES)

    remove_attrib = trim_attrib_paths(remove_attrib, remove_tags)
    create_attrib = trim_attrib_paths(create_attrib, create_tags)
    move_attrib = trim_attrib_move_paths(move_attrib, move_tags)

    remove_attrib = sorted(remove_attrib, key=lambda x: x.name)
    create_attrib = sorted(create_attrib, key=lambda x: x.name)
    move_attrib = sorted(move_attrib, key=lambda x: x.new_name)

    remove_attrib, create_attrib, move_attrib = _rename_elements(remove_attrib, create_attrib, move_attrib,
                                                                 from_version, to_version, 'attributes')

    remove_attrib = sorted(remove_attrib, key=lambda x: x.name)
    create_attrib = sorted(create_attrib, key=lambda x: x.name)
    move_attrib = sorted(move_attrib, key=lambda x: x.new_name)

    #Check again if we can now resolve ambiguouities
    resolve_ambiguouities(ambiguous_attrib,
                          remove_attrib,
                          create_attrib,
                          move_attrib,
                          remove_move=True,
                          tag_remove=remove_tags,
                          tag_move=move_tags)

    remove_attrib, create_attrib, move_attrib = _manual_resolution(ambiguous_attrib, remove_attrib, create_attrib,
                                                                   move_attrib, 'attributes')

    create_attrib = _create_attrib_elements(create_attrib, to_schema)
    remove_attrib = _add_warnings_on_remove(remove_attrib, 'attributes')

    conversion = FileConversion({
        'from_version': from_version,
        'to_version': to_version,
        'tag': {
            'remove': remove_tags,
            'create': create_tags,
            'move': cast('list[MoveAction| NormalizedMoveAction]', move_tags)
        },
        'attrib': {
            'remove': remove_attrib,
            'create': create_attrib,
            'move': cast('list[MoveAction| NormalizedMoveAction]', move_attrib)
        }
    })

    dump_conversion(conversion)

    if show:
        ctx.invoke(show_inp_conversion, from_version=from_version, to_version=to_version)

    return conversion


@inpxml.command('show-conversion')
@click.argument('from_version', type=str)
@click.argument('to_version', type=str)
def show_inp_conversion(from_version: str, to_version: str) -> None:
    """
    Show the actions for an already created conversion from FROM_VERSION to TO_VERSION

    FROM_VERSION is the file version of the initial input file
    TO_VERSION is the file version of the finale input file
    """
    try:
        conversion = load_conversion(from_version, to_version)
    except FileNotFoundError:
        echo.echo_critical(f'No conversion available between versions {from_version} to {to_version}')

    echo.echo_info(f'Actions performed for transforming from version {from_version} to {to_version}')
    echo_actions(conversion['tag']['remove'], header='The following tags are removed:')
    echo_actions(conversion['tag']['create'], header='The following tags are created:')
    echo_actions(conversion['tag']['move'], header='The following tags are moved:', ignore={'attrib', 'actual_path'})
    echo_actions(conversion['attrib']['remove'], header='The following attributes are removed:')
    echo_actions(conversion['attrib']['create'], header='The following attributes are created:')
    echo_actions(conversion['attrib']['move'],
                 header='The following attributes are moved:',
                 ignore={'attrib', 'actual_path'})



"""
This module is only here for backwards compatibility
"""
import warnings

warnings.warn('The io_fleurxml module was renamed to fleur_xml.\n'
              'Import is now masci_tools.io.fleur_xml', DeprecationWarning)

from .fleur_xml import *  #pylint: disable=unused-wildcard-import
